# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013-15, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

"""
CSV file based implementation of a record stream

FileRecordStream is class that can read and write .csv files that contain
records. The file has 3 header lines that contain, for each field, the name
(line 1), type (line 2), and a special indicator (line 3). The special indicator
can indicate that the field specifies a reset, is a sequence ID, or is a
timestamp for the record.

The header lines look like:

f1,f2,f3,....fN
int,string,datetime,bool,...
R,S,T,,,,....

The data lines are just comma separated values that match the types in the
second header line. The supported types are: int, float, string, bool, datetime

The format for datetime fields is yyyy-mm-dd hh:mm:ss.us
The 'us' component is microseconds.

When reading a file the FileRecordStream will automatically read the header line
and will figure out the type of each field and what are the timestamp, reset
and sequenceId fields (if any).

The FileRecordStream class supports the context manager ('with' statement )
protocol. That means you con do:

with FileRecordStream(filename) as f:
  ...
  ...

When the control exits the 'with' block the file will be closed automatically.
You may still call the .close() method at any point (even multiple times).

The FileRecordStream also supports the iteration protocol so you may read its
contents using a for loop:

  for r in f:
    print r
"""

import os
import csv
import copy
import json

from nupic.data.fieldmeta import FieldMetaInfo, FieldMetaType, FieldMetaSpecial
from nupic.data import SENTINEL_VALUE_FOR_MISSING_DATA
from nupic.data.record_stream import RecordStreamIface
from nupic.data.utils import (intOrNone, floatOrNone, parseBool, parseTimestamp,
    serializeTimestamp, serializeTimestampNoMS, escape, unescape, parseSdr,
    serializeSdr, parseStringList, stripList)



class FileRecordStream(RecordStreamIface):
  """ CSV file based RecordStream implementation
  """

  # Private: number of header rows (field names, types, special)
  _NUM_HEADER_ROWS = 3

  # Private: file mode for opening file for writing
  _FILE_WRITE_MODE = 'w'

  # Private: file mode for opening file for reading
  _FILE_READ_MODE = 'r'


  def __init__(self, streamID, write=False, fields=None, missingValues=None,
               bookmark=None, includeMS=True, firstRecord=None):
    """
    streamID:
        CSV file name, input or output
    write:
        True or False, open for writing if True
    fields:
        a list of nupic.data.fieldmeta.FieldMetaInfo field descriptors, only
        applicable when write==True
    missingValues:
        what missing values should be replaced with?
    bookmark:
        a reference to the previous reader, if passed in, the records will be
        returned starting from the point where bookmark was requested. Either
        bookmark or firstRecord can be specified, not both. If bookmark is used,
        then firstRecord MUST be None.
    includeMS:
        If false, the microseconds portion is not included in the
        generated output file timestamp fields. This makes it compatible
        with reading in from Excel.
    firstRecord:
        0-based index of the first record to start reading from. Either bookmark
        or firstRecord can be specified, not both. If bookmark is used, then
        firstRecord MUST be None.

    Each field is a 3-tuple (name, type, special or FieldMetaSpecial.none)

    The name is the name of the field. The type is one of the constants in
    `FieldMetaType`. The special is one of the `FieldMetaSpecial` values
    that designate their field as the sequenceId, reset, timestamp, or category.
    With exception of multiple categories, there can be at most one of each.
    There may be multiple fields of type datetime, but no more than one of them
    may be the timestamp field (FieldMetaSpecial.timestamp). The sequence id
    field must be either a string or an int. The reset field must be an int (and
    must contain 0 or 1).

    The category field must be an int or space-separated list of ints, where
    the former represents single-label classification and the latter is for
    multi-label classification (e.g. "1 3 4" designates a record for labels 1,
    3, and 4). The number of categories is allowed to vary record to record;
    sensor regions represent non-categories with -1, thus the category values
    must be >= 0.

    The FileRecordStream iterates over the field names, types and specials and
    stores the information.
    """
    super(FileRecordStream, self).__init__()

    # Only bookmark or firstRow can be specified, not both
    if bookmark is not None and firstRecord is not None:
      raise RuntimeError(
          "Only bookmark or firstRecord can be specified, not both")

    if fields is None:
      fields = []
    if missingValues is None:
      missingValues = ['']

    # We'll be operating on csvs with arbitrarily long fields
    size = 2**27
    csv.field_size_limit(size)

    self._filename = streamID
    # We can't guarantee what system files are coming from, use universal
    # newlines
    self._write = write
    self._mode = self._FILE_WRITE_MODE if write else self._FILE_READ_MODE
    self._file = open(self._filename, self._mode)
    self._sequences = set()
    self.rewindAtEOF = False

    if write:
      assert fields is not None
      assert isinstance(fields, (tuple, list))
      # Verify all fields are 3-tuple
      assert all(isinstance(f, (tuple, FieldMetaInfo)) and len(f) == 3
                 for f in fields)
      names, types, specials = zip(*fields)
      self._writer = csv.writer(self._file)
    else:
      # Read header lines
      self._reader = csv.reader(self._file, dialect="excel")
      try:
        names = [n.strip() for n in self._reader.next()]
      except:
        raise Exception('The header line of the file %s contained a NULL byte' \
                        % self._filename)
      types = [t.strip() for t in self._reader.next()]
      specials = [s.strip() for s in self._reader.next()]

      # If there are no specials, this means there was a blank line
      if len(specials) == 0:
        specials=[""]

    if not len(names) == len(types) == len(specials):
      raise Exception('Invalid file format: different number of fields '
                      'in the header rows of file %s (%d, %d, %d)' %
                      (streamID, len(names), len(types), len(specials)))

    # Verify standard file format
    for t in types:
      if not FieldMetaType.isValid(t):
        raise Exception('Invalid file format for "%s" - field type "%s" '
                        'not a valid FieldMetaType' % (self._filename, t,))

    for s in specials:
      if not FieldMetaSpecial.isValid(s):
        raise Exception('Invalid file format. \'%s\' is not a valid special '
                        'flag' % s)

    self._fields = [FieldMetaInfo(*attrs)
                    for attrs in zip(names, types, specials)]
    self._fieldCount = len(self._fields)

    # Keep track on how many records have been read/written
    self._recordCount = 0

    self._timeStampIdx = (specials.index(FieldMetaSpecial.timestamp)
                          if FieldMetaSpecial.timestamp in specials else None)
    self._resetIdx = (specials.index(FieldMetaSpecial.reset)
                      if FieldMetaSpecial.reset in specials else None)
    self._sequenceIdIdx = (specials.index(FieldMetaSpecial.sequence)
                           if FieldMetaSpecial.sequence in specials else None)
    self._categoryIdx = (specials.index(FieldMetaSpecial.category)
                         if FieldMetaSpecial.category in specials else None)
    self._learningIdx = (specials.index(FieldMetaSpecial.learning)
                         if FieldMetaSpecial.learning in specials else None)

    # keep track of the current sequence
    self._currSequence = None
    self._currTime = None

    if self._timeStampIdx:
      assert types[self._timeStampIdx] == FieldMetaType.datetime
    if self._sequenceIdIdx:
      assert types[self._sequenceIdIdx] in (FieldMetaType.string,
                                            FieldMetaType.integer)
    if self._resetIdx:
      assert types[self._resetIdx] == FieldMetaType.integer
    if self._categoryIdx:
      assert types[self._categoryIdx] in (FieldMetaType.list,
                                          FieldMetaType.integer)
    if self._learningIdx:
      assert types[self._learningIdx] == FieldMetaType.integer

    # Convert the types to the actual types in order to convert the strings
    if self._mode == self._FILE_READ_MODE:
      m = {FieldMetaType.integer: intOrNone,
           FieldMetaType.float: floatOrNone,
           FieldMetaType.boolean: parseBool,
           FieldMetaType.string: unescape,
           FieldMetaType.datetime: parseTimestamp,
           FieldMetaType.sdr: parseSdr,
           FieldMetaType.list: parseStringList}
    else:
      if includeMS:
        datetimeFunc = serializeTimestamp
      else:
        datetimeFunc = serializeTimestampNoMS
      m = {FieldMetaType.integer: str,
           FieldMetaType.float: str,
           FieldMetaType.string: escape,
           FieldMetaType.boolean: str,
           FieldMetaType.datetime: datetimeFunc,
           FieldMetaType.sdr: serializeSdr,
           FieldMetaType.list: stripList}

    self._adapters = [m[t] for t in types]

    self._missingValues = missingValues

    #
    # If the bookmark is set, we need to skip over first N records
    #
    if bookmark is not None:
      rowsToSkip = self._getStartRow(bookmark)
    elif firstRecord is not None:
      rowsToSkip = firstRecord
    else:
      rowsToSkip = 0

    while rowsToSkip > 0:
      self.next()
      rowsToSkip -= 1


    # Dictionary to store record statistics (min and max of scalars for now)
    self._stats = None


  def __getstate__(self):
    d = dict()
    d.update(self.__dict__)
    del d['_reader']
    del d['_file']
    return d


  def __setstate__(self, state):
    self.__dict__ = state
    self._file = None
    self._reader = None
    self.rewind()


  def close(self):
    if self._file is not None:
      self._file.close()
      self._file = None


  def rewind(self):
    """Put us back at the beginning of the file again)
    """

    # Superclass rewind
    super(FileRecordStream, self).rewind()

    self.close()
    self._file = open(self._filename, self._mode)
    self._reader = csv.reader(self._file, dialect="excel")

    # Skip header rows
    self._reader.next()
    self._reader.next()
    self._reader.next()

    # Reset record count, etc.
    self._recordCount = 0


  def getNextRecord(self, useCache=True):
    """ Returns next available data record from the file.

    retval: a data row (a list or tuple) if available; None, if no more records
             in the table (End of Stream - EOS); empty sequence (list or tuple)
             when timing out while waiting for the next record.
    """
    assert self._file is not None
    assert self._mode == self._FILE_READ_MODE

    # Read the line
    try:
      line = self._reader.next()

    except StopIteration:
      if self.rewindAtEOF:
        if self._recordCount == 0:
          raise Exception("The source configured to reset at EOF but "
                          "'%s' appears to be empty" % self._filename)
        self.rewind()
        line = self._reader.next()

      else:
        return None

    # Keep score of how many records were read
    self._recordCount += 1

    # Split the line to text fields and convert each text field to a Python
    # object if value is missing (empty string) encode appropriately for
    # upstream consumers in the case of numeric types, this means replacing
    # missing data with a sentinel value for string type, we can leave the empty
    # string in place
    record = []
    for i, f in enumerate(line):
      #print "DEBUG: Evaluating field @ index %s: %r" % (i, f)
      #sys.stdout.flush()
      if f in self._missingValues:
        record.append(SENTINEL_VALUE_FOR_MISSING_DATA)
      else:
        # either there is valid data, or the field is string type,
        # in which case the adapter does the right thing by default
        record.append(self._adapters[i](f))

    return record


  def getRecordsRange(self, bookmark=None, range=None):
    """ Returns a range of records, starting from the bookmark. If 'bookmark'
    is None, then records read from the first available. If 'range' is
    None, all available records will be returned (caution: this could be
    a lot of records and require a lot of memory).
    """

    raise Exception('getRecordsRange() is not supported for the file storage')


  def getLastRecords(self, numRecords):
    """ Returns a tuple (successCode, recordsArray), where
        successCode - if the stream had enough records to return, True/False
        recordsArray - an array of last numRecords records available when
                       the call was made. Records appended while in the
                       getLastRecords will be not returned until the next
                       call to either getNextRecord() or getLastRecords()
    """

    raise Exception('getLastRecords() is not supported for the file storage')


  def removeOldData(self):
    raise Exception('removeOldData is not supported in this class.')


  def appendRecord(self, record, inputBookmark=None):
    """ Saves the record in the underlying csv file.

        record: a list of Python objects that will be string-ified

        Returns: nothing
    """

    # input bookmark is not applicable in case of a file storage
    inputBookmark = inputBookmark

    assert self._file is not None
    assert self._mode == self._FILE_WRITE_MODE
    assert isinstance(record, (list, tuple)), \
      "unexpected record type: " + repr(type(record))

    assert len(record) == self._fieldCount, \
      "len(record): %s, fieldCount: %s" % (len(record), self._fieldCount)

    # Write header if needed
    if self._recordCount == 0:
      # Write the header
      names, types, specials = zip(*self.getFields())
      for line in names, types, specials:
        self._writer.writerow(line)

    # Keep track of sequences, make sure time flows forward
    self._updateSequenceInfo(record)

    line = [self._adapters[i](f) for i, f in enumerate(record)]

    self._writer.writerow(line)
    self._recordCount += 1


  def appendRecords(self, records, inputRef=None, progressCB=None):
    """ Saves multiple records in the underlying storage.

        Params: records - array of records as in 'appendRecord'
                inputRef - reference to the corresponding input (not applicable
                  in case of a file storage)
                progressCB - callback to report progress

        Returns: nothing

    """

    # input ref is not applicable in case of a file storage
    inputRef = inputRef

    for record in records:
      self.appendRecord(record, None)
      if progressCB is not None:
        progressCB()


  def getBookmark(self):
    """ Returns an anchor to the current position in the data. Passing this
    anchor to a constructor makes the current position to be the first
    returned record.
    """

    if self._write and self._recordCount==0:
      return None

    rowDict = dict(filepath=os.path.realpath(self._filename),
                   currentRow=self._recordCount)
    return json.dumps(rowDict)


  def recordsExistAfter(self, bookmark):
    """Returns True iff there are records left after the  bookmark."""
    return (self.getDataRowCount() - self.getNextRecordIdx()) > 0


  def seekFromEnd(self, numRecords):
    """Seeks to numRecords from the end and returns a bookmark to the new
    position.
    """
    self._file.seek(self._getTotalLineCount() - numRecords)
    return self.getBookmark()


  def setAutoRewind(self, autoRewind):
    """
    Controls whether getNext() should automatically rewind the source when EOF
    is reached.

    autoRewind: True = getNext() will automatically rewind the source on EOF;
                False = getNext() will not automatically rewind the source
                on EOF
    """
    self.rewindAtEOF = autoRewind


  def getStats(self):
    """ Parse the file using dedicated reader and collect fields stats. Never
    called if user of FileRecordStream does not invoke getStats method.

    Returns: a dictionary of stats. In the current implementation, min and max
             fields are supported. Example of the return dictionary is:

             {
               'min' : [f1_min, f2_min, None, None, fn_min],
               'max' : [f1_max, f2_max, None, None, fn_max]
             }

             (where fx_min/fx_max are set for scalar fields, or None if not)

    """

    # Collect stats only once per File object, use fresh csv iterator
    # to keep the next() method returning sequential records no matter when
    # caller asks for stats
    if self._stats == None:
      # Stats are only available when reading csv file
      assert self._mode == self._FILE_READ_MODE

      inFile = open(self._filename, self._FILE_READ_MODE)

      # Create a new reader; read names, types, specials
      reader = csv.reader(inFile, dialect="excel")
      names = [n.strip() for n in reader.next()]
      types = [t.strip() for t in reader.next()]
      # Skip over specials
      reader.next()

      # Initialize stats to all None
      self._stats = dict()
      self._stats['min'] = []
      self._stats['max'] = []

      for i in xrange(len(names)):
        self._stats['min'].append(None)
        self._stats['max'].append(None)

      # Read the file, collect stats
      while True:
        try:
          line = reader.next()
          for i, f in enumerate(line):
            if (len(types) > i and
                types[i] in [FieldMetaType.integer, FieldMetaType.float] and
                f not in self._missingValues):
              value = self._adapters[i](f)
              if self._stats['max'][i] == None or \
                 self._stats['max'][i] < value:
                self._stats['max'][i] = value
              if self._stats['min'][i] == None or \
                 self._stats['min'][i] > value:
                self._stats['min'][i] = value

        except StopIteration:
          break

    return self._stats


  def clearStats(self):
    """ Resets stats collected so far.
    """
    self._stats = None


  def getError(self):
    """ Returns errors saved in the stream.
    """
    # CSV file version does not provide storage for the error information
    return None


  def setError(self, error):
    """ Saves specified error in the stream.
    """
    # CSV file version does not provide storage for the error information
    return


  def isCompleted(self):
    """ Returns True if all records are already in the stream or False
    if more records is expected.
    """
    # CSV file is always considered completed
    return True


  def setCompleted(self, completed=True):
    """ Marks the stream completed (True or False)
    """
    # CSV file is always considered completed, nothing to do
    return


  def getFieldNames(self):
    """ Returns an array of field names associated with the data.
    """
    return [f.name for f in self._fields]


  def getFields(self):
    """ Returns a sequence of nupic.data.fieldmeta.FieldMetaInfo
    name/type/special tuples for each field in the stream.
    """
    if self._fields is None:
      return None
    else:
      return copy.copy(self._fields)


  def _updateSequenceInfo(self, r):
    """Keep track of sequence and make sure time goes forward

    Check if the current record is the beginning of a new sequence
    A new sequence starts in 2 cases:

    1. The sequence id changed (if there is a sequence id field)
    2. The reset field is 1 (if there is a reset field)

    Note that if there is no sequenceId field or resetId field then the entire
    dataset is technically one big sequence. The function will not return True
    for the first record in this case. This is Ok because it is important to
    detect new sequences only when there are multiple sequences in the file.
    """

    # Get current sequence id (if any)
    newSequence = False
    sequenceId = (r[self._sequenceIdIdx]
                  if self._sequenceIdIdx is not None else None)
    if sequenceId != self._currSequence:
      # verify that the new sequence didn't show up before
      if sequenceId in self._sequences:
        raise Exception('Broken sequence: %s, record: %s' % \
                        (sequenceId, r))

      # add the finished sequence to the set of sequence
      self._sequences.add(self._currSequence)
      self._currSequence = sequenceId

      # Verify that the reset is consistent (if there is one)
      if self._resetIdx:
        assert r[self._resetIdx] == 1
      newSequence = True

    else:
      # Check the reset
      reset = False
      if self._resetIdx:
        reset = r[self._resetIdx]
        if reset == 1:
          newSequence = True

    # If it's still the same old sequence make sure the time flows forward
    if not newSequence:
      if self._timeStampIdx and self._currTime is not None:
        t = r[self._timeStampIdx]
        if t < self._currTime:
          raise Exception('No time travel. Early timestamp for record: %s' % r)

    if self._timeStampIdx:
      self._currTime = r[self._timeStampIdx]


  def _getStartRow(self, bookmark):
    """ Extracts start row from the bookmark information
    """
    bookMarkDict = json.loads(bookmark)

    realpath = os.path.realpath(self._filename)

    bookMarkFile = bookMarkDict.get('filepath', None)

    if bookMarkFile != realpath:
      print ("Ignoring bookmark due to mismatch between File's "
             "filename realpath vs. bookmark; realpath: %r; bookmark: %r") % (
        realpath, bookMarkDict)
      return 0
    else:
      return bookMarkDict['currentRow']


  def _getTotalLineCount(self):
    """ Returns:  count of ALL lines in dataset, including header lines
    """
    # Flush the file before we open it again to count lines
    if self._mode == self._FILE_WRITE_MODE:
      self._file.flush()
    return sum(1 for line in open(self._filename, self._FILE_READ_MODE))


  def getNextRecordIdx(self):
    """Returns the index of the record that will be read next from
    getNextRecord()
    """
    return self._recordCount


  def getDataRowCount(self):
    """
    Returns:  count of data rows in dataset (excluding header lines)
    """
    numLines = self._getTotalLineCount()

    if numLines == 0:
      # this may be the case in a file opened for write before the
      # header rows are written out
      assert self._mode == self._FILE_WRITE_MODE and self._recordCount == 0
      numDataRows = 0
    else:
      numDataRows = numLines - self._NUM_HEADER_ROWS

    assert numDataRows >= 0

    return numDataRows


  def setTimeout(self, timeout):
    """ Set the read timeout """
    pass


  def flush(self):
    if self._file is not None:
      self._file.flush()


  def __enter__(self):
    """Context guard - enter

    Just return the object
    """
    return self


  def __exit__(self, yupe, value, traceback):
    """Context guard - exit

    Ensures that the file is always closed at the end of the 'with' block.
    Lets exceptions propagate.
    """
    self.close()


  def __iter__(self):
    """Support for the iterator protocol. Return itself"""
    return self


  def next(self):
    """Implement the iterator protocol """
    record = self.getNextRecord()
    if record is None:
      raise StopIteration

    return record




# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------



from abc import ABCMeta, abstractmethod

import numbers
import copy
import random
import numpy as np

from nupic.data.fieldmeta import FieldMetaType
import nupic.math.roc_utils as roc
from nupic.data import SENTINEL_VALUE_FOR_MISSING_DATA
from nupic.frameworks.opf.opfutils import InferenceType
from nupic.utils import MovingAverage

from collections import deque
from operator import itemgetter
from safe_interpreter import SafeInterpreter
from io import BytesIO, StringIO
from functools import partial

###############################################################################
# Public Metric specification class
###############################################################################

class MetricSpec(object):
  """ This class represents a single Metrics specification in the TaskControl
  block
  """

  _LABEL_SEPARATOR = ":"

  def __init__(self, metric, inferenceElement, field=None, params=None):
    """
    metric:           A metric type name that identifies which metrics module is
                      to be constructed by the metrics factory method
                      opf.metrics.getModule(); e.g., "rmse"

    inferenceElement: Some inference types (such as classification), can output
                      more than one type of inference (i.e. the predicted class
                      AND the predicted next step). This field specifies which
                      of these inferences to compute the metrics on

    field:            Field name on which this metric is to be collected
    params:           Custom parameters dict for the metrics module's constructor
    """

    self.metric = metric
    self.inferenceElement = inferenceElement
    self.field = field
    self.params = params
    return

  def __repr__(self):
    return "%s(metric=%r, inferenceElement=%r, field=%r, params=%r)" \
                      % (self.__class__.__name__,
                         self.metric,
                         self.inferenceElement,
                         self.field,
                         self.params)

  def getLabel(self, inferenceType=None):
    """ Helper method that generates a unique label
    for a MetricSpec / InferenceType pair. The label is formatted
    as follows:

      <predictionKind>:<metric type>:(paramName=value)*:field=<fieldname>

    For example:
      classification:aae:paramA=10.2:paramB=20:window=100:field=pounds
    """
    result = []
    if inferenceType is not None:
      result.append(InferenceType.getLabel(inferenceType))
    result.append(self.inferenceElement)
    result.append(self.metric)

    params = self.params
    if params is not None:

      sortedParams= params.keys()
      sortedParams.sort()
      for param in sortedParams:
        # Don't include the customFuncSource - it is too long an unwieldy
        if param in ('customFuncSource', 'customFuncDef', 'customExpr'):
          continue
        value = params[param]
        if isinstance(value, str):
          result.extend(["%s='%s'"% (param, value)])
        else:
          result.extend(["%s=%s"% (param, value)])

    if self.field:
      result.append("field=%s"% (self.field) )

    return self._LABEL_SEPARATOR.join(result)

  @classmethod
  def getInferenceTypeFromLabel(cls, label):
    """ Extracts the PredicitonKind (temporal vs. nontemporal) from the given
    metric label

    Parameters:
    -----------------------------------------------------------------------
    label:      A label (string) for a metric spec generated by getMetricLabel
                (above)

    Returns:   An InferenceType value
    """
    infType, _, _= label.partition(cls._LABEL_SEPARATOR)

    if not InferenceType.validate(infType):
      return None

    return infType



def getModule(metricSpec):
  """
      factory method to return an appropriate MetricsIface-based module
      args:
          metricSpec - an instance of MetricSpec.
          metricSpec.metric must be one of:
              rmse (root-mean-square error)
              aae (average absolute error)
              acc (accuracy, for enumerated types)
      return:
          an appropriate Metric module
  """

  metricName = metricSpec.metric

  if metricName == 'rmse':
    return MetricRMSE(metricSpec)
  if metricName == 'nrmse':
    return MetricNRMSE(metricSpec)
  elif metricName == 'aae':
    return MetricAAE(metricSpec)
  elif metricName == 'acc':
    return MetricAccuracy(metricSpec)
  elif metricName == 'avg_err':
    return MetricAveError(metricSpec)
  elif metricName == 'trivial':
    return MetricTrivial(metricSpec)
  elif metricName == 'two_gram':
    return MetricTwoGram(metricSpec)
  elif metricName == 'moving_mean':
    return MetricMovingMean(metricSpec)
  elif metricName == 'moving_mode':
    return MetricMovingMode(metricSpec)
  elif metricName == 'neg_auc':
    return MetricNegAUC(metricSpec)
  elif metricName == 'custom_error_metric':
    return CustomErrorMetric(metricSpec)
  elif metricName == 'multiStep':
    return MetricMultiStep(metricSpec)
  elif metricName == 'multiStepProbability':
    return MetricMultiStepProbability(metricSpec)
  elif metricName == 'ms_aae':
    return MetricMultiStepAAE(metricSpec)
  elif metricName == 'ms_avg_err':
    return MetricMultiStepAveError(metricSpec)
  elif metricName == 'passThruPrediction':
    return MetricPassThruPrediction(metricSpec)
  elif metricName == 'altMAPE':
    return MetricAltMAPE(metricSpec)
  elif metricName == 'MAPE':
    return MetricMAPE(metricSpec)
  elif metricName == 'multi':
    return MetricMulti(metricSpec)
  elif metricName == 'negativeLogLikelihood':
    return MetricNegativeLogLikelihood(metricSpec)
  else:
    raise Exception("Unsupported metric type: %s" % metricName)

################################################################################
#               Helper Methods and Classes                                    #
################################################################################

class _MovingMode(object):
  """ Helper class for computing windowed moving
  mode of arbitrary values """

  def __init__(self, windowSize = None):
    """
    Parameters:
    -----------------------------------------------------------------------
    windowSize:             The number of values that are used to compute the
                            moving average
    """
    self._windowSize = windowSize
    self._countDict = dict()
    self._history = deque([])


  def __call__(self, value):

    if len(self._countDict) == 0:
      pred = ""
    else:
      pred = max(self._countDict.items(), key = itemgetter(1))[0]

    # Update count dict and history buffer
    self._history.appendleft(value)

    if not value in self._countDict:
      self._countDict[value] = 0
    self._countDict[value] += 1

    if len(self._history) > self._windowSize:
      removeElem = self._history.pop()
      self._countDict[removeElem] -= 1
      assert(self._countDict[removeElem] > -1)

    return pred



def _isNumber(value):
  return isinstance(value, (numbers.Number, np.number))



class MetricsIface(object):
  """
  A Metrics module compares a prediction Y to corresponding ground truth X and returns a single
  measure representing the "goodness" of the prediction. It is up to the implementation to
  determine how this comparison is made.

  """

  __metaclass__ = ABCMeta

  @abstractmethod
  def __init__(self, metricSpec):
    """
        instantiate a MetricsIface-based module.
        args:
            metricSpec is an instance of MetricSpec

    """

  @abstractmethod
  def addInstance(self, groundTruth, prediction, record = None, result = None):
    """ add one instance consisting of ground truth and a prediction.

      Parameters:
      -----------------------------------------------------------------------
      groundTruth:
        The actual measured value at the current timestep
      prediction:
        The value predicted by the network at the current timestep

      groundTruthEncoding:
        The binary encoding of the groundTruth value (as a numpy array). Right
        now this is only used by CLA networks
      predictionEncoding:
        The binary encoding of the prediction value (as a numpy array). Right
        now this is only used by CLA networks

      result:
        An ModelResult class (see opfutils.py)

        return:
            The average error as computed over the metric's window size
    """

  @abstractmethod
  def getMetric(self):
    """
        return:
            {value : <current measurement>, "stats" : {<stat> : <value> ...}}
            metric name is defined by the MetricIface implementation. stats is expected to contain further
                information relevant to the given metric, for example the number of timesteps represented in
                the current measurement. all stats are implementation defined, and "stats" can be None

    """



class AggregateMetric(MetricsIface):
  """
      Partial implementation of Metrics Interface for metrics that
      accumulate an error and compute an aggregate score, potentially
      over some window of previous data. This is a convenience class that
      can serve as the base class for a wide variety of metrics
  """
  ___metaclass__ = ABCMeta

  #FIXME @abstractmethod - this should be marked abstract method and required to be implemented
  def accumulate(self, groundTruth, prediction, accumulatedError, historyBuffer, result):
    """
        Updates the accumulated error given the prediction and the
        ground truth.

        groundTruth: Actual value that is observed for the current timestep

        prediction: Value predicted by the network for the given timestep

        accumulatedError: The total accumulated score from the previous
          predictions (possibly over some finite window)

        historyBuffer: A buffer of the last <self.window> ground truth values
            that have been observed.

            If historyBuffer = None,  it means that no history is being kept.

        result: An ModelResult class (see opfutils.py), used for advanced
          metric calculation (e.g., MetricNegativeLogLikelihood)

          retval:
            The new accumulated error. That is:
            self.accumulatedError = self.accumulate(groundTruth, predictions, accumulatedError)

            historyBuffer should also be updated in this method.
            self.spec.params["window"] indicates the maximum size of the window
    """

  #FIXME @abstractmethod - this should be marked abstract method and required to be implemented
  def aggregate(self, accumulatedError, historyBuffer, steps):
    """
        Updates the final aggregated score error given the prediction and the
        ground truth.

        accumulatedError: The total accumulated score from the previous
          predictions (possibly over some finite window)

        historyBuffer: A buffer of the last <self.window> ground truth values
            that have been observed.

            If historyBuffer = None,  it means that no history is being kept.

        steps: The total number of (groundTruth, prediction) pairs that have
        been passed to the metric. This does not include pairs where
        the groundTruth = SENTINEL_VALUE_FOR_MISSING_DATA

          retval:
            The new aggregate (final) error measure.
    """

  def __init__(self, metricSpec):
    """ Initialize this metric

    If the params contains the key 'errorMetric', then that is the name of
    another metric to which we will pass a modified groundTruth and prediction
    to from our addInstance() method. For example, we may compute a moving mean
    on the groundTruth and then pass that to the AbsoluteAveError metric
    """

    # Init default member variables
    self.id = None
    self.verbosity = 0
    self.window = -1
    self.history = None
    self.accumulatedError = 0
    self.aggregateError = None
    self.steps = 0
    self.spec = metricSpec
    self.disabled = False

    # Number of steps ahead we are trying to predict. This is a list of
    #  prediction steps are processing
    self._predictionSteps = [0]

    # Where we store the ground truth history
    self._groundTruthHistory = deque([])

    # The instances of another metric to which we will pass a possibly modified
    #  groundTruth and prediction to from addInstance(). There is one instance
    #  for each step present in self._predictionSteps
    self._subErrorMetrics = None

    # The maximum number of records to process. After this many records have
    #  been processed, the metric value never changes. This can be used
    #  as the optimization metric for swarming, while having another metric without
    #  the maxRecords limit to get an idea as to how well a production model
    #  would do on the remaining data
    self._maxRecords = None

    # Parse the metric's parameters
    if metricSpec is not None and metricSpec.params is not None:
      
      self.id = metricSpec.params.get('id', None)
      self._predictionSteps = metricSpec.params.get('steps', [0])
      # Make sure _predictionSteps is a list
      if not hasattr(self._predictionSteps, '__iter__'):
        self._predictionSteps = [self._predictionSteps]
        
      self.verbosity = metricSpec.params.get('verbosity', 0)
      self._maxRecords = metricSpec.params.get('maxRecords', None)

      # Get the metric window size
      if 'window' in metricSpec.params:
        assert metricSpec.params['window'] >= 1
        self.history = deque([])
        self.window = metricSpec.params['window']

      # Get the name of the sub-metric to chain to from addInstance()
      if 'errorMetric' in metricSpec.params:
        self._subErrorMetrics = []
        for step in self._predictionSteps:
          subSpec = copy.deepcopy(metricSpec)
          # Do all ground truth shifting before we pass onto the sub-metric
          subSpec.params.pop('steps', None)
          subSpec.params.pop('errorMetric')
          subSpec.metric = metricSpec.params['errorMetric']
          self._subErrorMetrics.append(getModule(subSpec))



  def _getShiftedGroundTruth(self, groundTruth):
    """ Utility function that saves the passed in groundTruth into a local
    history buffer, and returns the groundTruth from self._predictionSteps ago, 
    where self._predictionSteps is defined by the 'steps' parameter.
    This can be called from the beginning of a derived class's addInstance()
    before it passes groundTruth and prediction onto accumulate().
    """

    # Save this ground truth into our input history
    self._groundTruthHistory.append(groundTruth)

    # This is only supported when _predictionSteps has one item in it
    assert (len(self._predictionSteps) == 1)
    # Return the one from N steps ago
    if len(self._groundTruthHistory) > self._predictionSteps[0]:
      return self._groundTruthHistory.popleft()
    else:
      if hasattr(groundTruth, '__iter__'):
        return [None] * len(groundTruth)
      else:
        return None


  def addInstance(self, groundTruth, prediction, record = None, result = None):

    # This base class does not support time shifting the ground truth or a
    #  subErrorMetric.
    assert (len(self._predictionSteps) == 1)
    assert self._predictionSteps[0] == 0
    assert self._subErrorMetrics is None


    # If missing data,
    if groundTruth == SENTINEL_VALUE_FOR_MISSING_DATA or prediction is None:
      return self.aggregateError

    if self.verbosity > 0:
      print "groundTruth:\n%s\nPredictions:\n%s\n%s\n" % (groundTruth,
                                                prediction, self.getMetric())

    # Ignore if we've reached maxRecords
    if self._maxRecords is not None and self.steps >= self._maxRecords:
      return self.aggregateError

    # If there is a sub-metric, chain into it's addInstance
    # Accumulate the error
    self.accumulatedError = self.accumulate(groundTruth, prediction,
                                            self.accumulatedError, self.history, result)

    self.steps += 1
    return self._compute()

  def getMetric(self):
    return {'value': self.aggregateError, "stats" : {"steps" : self.steps}}

  def _compute(self):
    self.aggregateError = self.aggregate(self.accumulatedError, self.history,
                                         self.steps)
    return self.aggregateError


class MetricNegativeLogLikelihood(AggregateMetric):
  """
      computes negative log-likelihood. Likelihood is the predicted probability of
      the true data from a model. It is more powerful than metrics that only considers
      the single best prediction (e.g. MSE) as it considers the entire probability
      distribution predicted by a model.

      It is more appropriate to use likelihood as the error metric when multiple
      predictions are possible.
  """
  def accumulate(self, groundTruth, prediction, accumulatedError, historyBuffer, result):
    bucketll = result.inferences['multiStepBucketLikelihoods']
    bucketIdxTruth = result.classifierInput.bucketIndex

    if bucketIdxTruth is not None:
      # a manually set minimum prediction probability so that the log(LL) doesn't blow up
      minProb = 0.00001
      negLL = 0
      for step in bucketll.keys():
        outOfBucketProb = 1 - sum(bucketll[step].values())
        if bucketIdxTruth in bucketll[step].keys():
          prob = bucketll[step][bucketIdxTruth]
        else:
          prob = outOfBucketProb

        if prob < minProb:
          prob = minProb
        negLL -= np.log(prob)

      accumulatedError += negLL

      if historyBuffer is not None:
        historyBuffer.append(negLL)
        if len(historyBuffer) > self.spec.params["window"]:
          accumulatedError -= historyBuffer.popleft()

    return accumulatedError

  def aggregate(self, accumulatedError, historyBuffer, steps):
    n = steps
    if historyBuffer is not None:
      n = len(historyBuffer)

    return accumulatedError / float(n)


class MetricRMSE(AggregateMetric):
  """
      computes root-mean-square error
  """
  def accumulate(self, groundTruth, prediction, accumulatedError, historyBuffer, result = None):
    error = (groundTruth - prediction)**2
    accumulatedError += error

    if historyBuffer is not None:
      historyBuffer.append(error)
      if len(historyBuffer) > self.spec.params["window"] :
        accumulatedError -= historyBuffer.popleft()

    return accumulatedError

  def aggregate(self, accumulatedError, historyBuffer, steps):
    n = steps
    if historyBuffer is not None:
      n = len(historyBuffer)

    return np.sqrt(accumulatedError / float(n))



class MetricNRMSE(MetricRMSE):
  """computes normalized root-mean-square error"""
  def __init__(self, *args, **kwargs):
    super(MetricNRMSE, self).__init__(*args, **kwargs)
    self.groundTruths = []

  def accumulate(self, groundTruth, prediction, accumulatedError, historyBuffer, result = None):
    self.groundTruths.append(groundTruth)

    return super(MetricNRMSE, self).accumulate(groundTruth,
                                               prediction,
                                               accumulatedError,
                                               historyBuffer,
                                               result)

  def aggregate(self, accumulatedError, historyBuffer, steps):
    rmse = super(MetricNRMSE, self).aggregate(accumulatedError,
                                              historyBuffer,
                                              steps)
    denominator = np.std(self.groundTruths)
    return rmse / denominator if denominator > 0 else float("inf")



class MetricAAE(AggregateMetric):
  """
      computes average absolute error
  """
  def accumulate(self, groundTruth, prediction, accumulatedError, historyBuffer, result = None):
    error = abs(groundTruth - prediction)
    accumulatedError += error

    if historyBuffer is not None:
      historyBuffer.append(error)
      if len(historyBuffer) > self.spec.params["window"] :
        accumulatedError -= historyBuffer.popleft()

    return accumulatedError

  def aggregate(self, accumulatedError, historyBuffer, steps):
    n = steps
    if historyBuffer is not None:
      n = len(historyBuffer)

    return accumulatedError/ float(n)



class MetricAltMAPE(AggregateMetric):
  """
  computes the "Alternative" Mean Absolute Percent Error.

  A generic MAPE computes the percent error for each sample, and then gets
  an average. This can suffer from samples where the actual value is very small
  or zero - this one sample can drastically alter the mean.

  This metric on the other hand first computes the average of the actual values
  and the averages of the errors before dividing. This washes out the effects of
  a small number of samples with very small actual values.
  """

  def __init__(self, metricSpec):
    super(MetricAltMAPE, self).__init__(metricSpec)
    self._accumulatedGroundTruth = 0
    self._accumulatedError = 0

  def addInstance(self, groundTruth, prediction, record = None, result = None):

    # If missing data,
    if groundTruth == SENTINEL_VALUE_FOR_MISSING_DATA or prediction is None:
      return self.aggregateError


    # Compute absolute error
    error = abs(groundTruth - prediction)
    if self.verbosity > 0:
      print "MetricAltMAPE:\n  groundTruth: %s\n  Prediction: " \
            "%s\n  Error: %s" % (groundTruth, prediction, error)

    # Update the accumulated groundTruth and aggregate error
    if self.history is not None:
      self.history.append((groundTruth, error))
      if len(self.history) > self.spec.params["window"] :
        (oldGT, oldErr) = self.history.popleft()
        self._accumulatedGroundTruth -= oldGT
        self._accumulatedError -= oldErr

    self._accumulatedGroundTruth += abs(groundTruth)
    self._accumulatedError += error

    # Compute aggregate pct error
    if self._accumulatedGroundTruth > 0:
      self.aggregateError = 100.0 * self._accumulatedError / \
                              self._accumulatedGroundTruth
    else:
      self.aggregateError = 0

    if self.verbosity >= 1:
      print "  accumGT:", self._accumulatedGroundTruth
      print "  accumError:", self._accumulatedError
      print "  aggregateError:", self.aggregateError

    self.steps += 1
    return self.aggregateError



class MetricMAPE(AggregateMetric):
  """
  computes the "Classic" Mean Absolute Percent Error.

  This computes the percent error for each sample, and then gets
  an average. Note that this can suffer from samples where the actual value is 
  very small or zero - this one sample can drastically alter the mean. To 
  avoid this potential issue, use 'altMAPE' instead. 
  
  This metric is provided mainly as a convenience when comparing results against
  other investigations that have also used MAPE. 

  """

  def __init__(self, metricSpec):
    super(MetricMAPE, self).__init__(metricSpec)
    self._accumulatedPctError = 0

  def addInstance(self, groundTruth, prediction, record = None, result = None):

    # If missing data,
    if groundTruth == SENTINEL_VALUE_FOR_MISSING_DATA or prediction is None:
      return self.aggregateError


    # Compute absolute error
    if groundTruth != 0:
      pctError = float(abs(groundTruth - prediction))/groundTruth
    else:
      # Ignore this sample
      if self.verbosity > 0:
        print "Ignoring sample with groundTruth of 0"
      self.steps += 1
      return self.aggregateError
    
    if self.verbosity > 0:
      print "MetricMAPE:\n  groundTruth: %s\n  Prediction: " \
            "%s\n  Error: %s" % (groundTruth, prediction, pctError)

    # Update the accumulated groundTruth and aggregate error
    if self.history is not None:
      self.history.append(pctError)
      if len(self.history) > self.spec.params["window"] :
        (oldPctErr) = self.history.popleft()
        self._accumulatedPctError -= oldPctErr

    self._accumulatedPctError += pctError

    # Compute aggregate pct error
    self.aggregateError = 100.0 * self._accumulatedPctError / len(self.history)

    if self.verbosity >= 1:
      print "  accumPctError:", self._accumulatedPctError
      print "  aggregateError:", self.aggregateError

    self.steps += 1
    return self.aggregateError



class MetricPassThruPrediction(MetricsIface):
  """
  This is not a metric, but rather a facility for passing the predictions
  generated by a baseline metric through to the prediction output cache produced
  by a model.

  For example, if you wanted to see the predictions generated for the TwoGram
  metric, you would specify 'PassThruPredictions' as the 'errorMetric' parameter.

  This metric class simply takes the prediction and outputs that as the
  aggregateMetric value.
  """
  
  def __init__(self, metricSpec):
    self.spec = metricSpec
    self.window = metricSpec.params.get("window", 1)
    self.avg = MovingAverage(self.window)
    
    self.value = None
    
  def addInstance(self, groundTruth, prediction, record = None, result = None):
    """Compute and store metric value"""
    self.value = self.avg(prediction)
    
  def getMetric(self):
    """Return the metric value """
    return {"value": self.value}
    

  #def accumulate(self, groundTruth, prediction, accumulatedError, historyBuffer):
  #  # Simply return the prediction as the accumulated error
  #  return prediction
  #
  #def aggregate(self, accumulatedError, historyBuffer, steps):
  #  # Simply return the prediction as the aggregateError
  #  return accumulatedError



class MetricMovingMean(AggregateMetric):
  """
      computes error metric based on moving mean prediction
  """
  def __init__(self, metricSpec):

    # This metric assumes a default 'steps' of 1
    if not 'steps' in metricSpec.params:
      metricSpec.params['steps'] = 1

    super(MetricMovingMean, self).__init__(metricSpec)
    
    # Only supports 1 item in _predictionSteps
    assert (len(self._predictionSteps) == 1)

    self.mean_window = 10
    if metricSpec.params.has_key('mean_window'):
      assert metricSpec.params['mean_window'] >= 1
      self.mean_window = metricSpec.params['mean_window']

    # Construct moving average instance
    self._movingAverage = MovingAverage(self.mean_window)

  def getMetric(self):
    return self._subErrorMetrics[0].getMetric()

  def addInstance(self, groundTruth, prediction, record = None, result = None):

    # If missing data,
    if groundTruth == SENTINEL_VALUE_FOR_MISSING_DATA:
      return self._subErrorMetrics[0].aggregateError

    if self.verbosity > 0:
      print "groundTruth:\n%s\nPredictions:\n%s\n%s\n" % (groundTruth, prediction, self.getMetric())

    # Use ground truth from 'steps' steps ago as our most recent ground truth
    lastGT = self._getShiftedGroundTruth(groundTruth)
    if lastGT is None:
      return self._subErrorMetrics[0].aggregateError

    mean = self._movingAverage(lastGT)

    return self._subErrorMetrics[0].addInstance(groundTruth, mean, record)

def evalCustomErrorMetric(expr, prediction, groundTruth, tools):
  sandbox = SafeInterpreter(writer=StringIO())
  if isinstance(prediction, dict):
    sandbox.symtable['prediction'] = tools.mostLikely(prediction)
    sandbox.symtable['EXP'] = tools.expValue(prediction)
    sandbox.symtable['probabilityDistribution'] = prediction
  else:
    sandbox.symtable['prediction'] = prediction
  sandbox.symtable['groundTruth'] = groundTruth
  sandbox.symtable['tools'] = tools
  error = sandbox(expr)
  return error



class CustomErrorMetric(MetricsIface):
  """
    Custom Error Metric class that handles user defined error metrics
  """
  class CircularBuffer():
    """
      implementation of a fixed size constant random access circular buffer
    """
    def __init__(self,length):
      #Create an array to back the buffer
      #If the length<0 create a zero length array
      self.data = [None for i in range(max(length,0))]
      self.elements = 0
      self.index = 0
      self.dataLength = length

    def getItem(self,n):
      #Get item from n steps back
      if n >= self.elements or (n >= self.dataLength and not self.dataLength < 0):
        assert  False,"Trying to access data not in the stored window"
        return None
      if self.dataLength>=0:
        getInd = (self.index-n-1)%min(self.elements,self.dataLength)
      else:
        getInd = (self.index-n-1)%self.elements
      return self.data[getInd]

    def pushToEnd(self,obj):
      ret = None
      #If storing everything simply append right to the list
      if(self.dataLength < 0 ):
        self.data.append(obj)
        self.index+=1
        self.elements+=1
        return None
      if(self.elements==self.dataLength):
        #pop last added element
        ret = self.data[self.index % self.dataLength]
      else:
        #else push new element and increment the element counter
        self.elements += 1
      self.data[self.index % self.dataLength] = obj
      self.index += 1
      return ret
    def __len__(self):
      return self.elements
  
  def __init__(self,metricSpec):
    self.metricSpec = metricSpec
    self.steps = 0
    self.error = 0
    self.averageError = None
    self.errorMatrix = None
    self.evalError = self.evalAbsErr
    self.errorWindow = 1
    self.storeWindow=-1
    self.userDataStore = dict()
    if "errorWindow" in metricSpec.params:
      self.errorWindow = metricSpec.params["errorWindow"]
      assert self.errorWindow  != 0 , "Window Size cannon be zero"
    if "storeWindow" in metricSpec.params:
      self.storeWindow = metricSpec.params["storeWindow"]
      assert self.storeWindow  != 0 , "Window Size cannon be zero"
    self.errorStore = self.CircularBuffer(self.errorWindow)
    self.recordStore = self.CircularBuffer(self.storeWindow)
    if "customExpr" in metricSpec.params:
      assert not "customFuncDef" in metricSpec.params
      assert not "customFuncSource" in metricSpec.params
      self.evalError = partial(evalCustomErrorMetric, metricSpec.params["customExpr"])
    elif "customFuncSource" in metricSpec.params:
      assert not "customFuncDef" in metricSpec.params
      assert not "customExpr" in metricSpec.params
      exec(metricSpec.params["customFuncSource"])
      #pull out defined function from locals
      self.evalError = locals()["getError"]
    elif "customFuncDef" in metricSpec.params:
      assert not "customFuncSource" in metricSpec.params
      assert not "customExpr" in metricSpec.params
      self.evalError = metricSpec.params["customFuncDef"]
  
  def getPrediction(self,n):
    #Get prediction from n steps ago
    return self.recordStore.getItem(n)["prediction"]
  
  def getFieldValue(self,n,field):
    #Get field value from record n steps ago
    record = self.recordStore.getItem(n)["record"]
    value = record[field]
    return value
  
  def getGroundTruth(self,n):
    #Get the groundTruth from n steps ago
    return self.recordStore.getItem(n)["groundTruth"]
  
  def getBufferLen(self):
    return len(self.recordStore)
  
  def storeData(self,name,obj):
    #Store custom user data
    self.userDataStore[name] = obj
  
  def getData(self,name):
    #Retrieve user data
    if name in self.userDataStore:
      return self.userDataStore[name]
    return None

  def mostLikely(self, pred):
    """ Helper function to return a scalar value representing the most
        likely outcome given a probability distribution
    """
    if len(pred) == 1:
      return pred.keys()[0]

    mostLikelyOutcome = None
    maxProbability = 0
    
    for prediction, probability in pred.items():
      if probability > maxProbability:
        mostLikelyOutcome = prediction
        maxProbability = probability

    return mostLikelyOutcome

  def expValue(self, pred):
    """ Helper function to return a scalar value representing the expected 
        value of a probability distribution
    """
    if len(pred) == 1:
      return pred.keys()[0]
    
    return sum([x*p for x,p in pred.items()])

  def evalAbsErr(self,pred,ground):
    return abs(pred-ground)

  def getMetric(self):
    return {'value': self.averageError, "stats" : {"steps" : self.steps}}

  def addInstance(self, groundTruth, prediction, record = None, result = None):

    #If missing data,
    if groundTruth == SENTINEL_VALUE_FOR_MISSING_DATA or prediction is None:
      return self.averageError
    
    self.recordStore.pushToEnd({"groundTruth":groundTruth,
                                         "prediction":prediction,"record":record})

    if isinstance(prediction, dict):
      assert not any(True for p in prediction if p is None), \
        "Invalid prediction of `None` in call to %s.addInstance()" % \
          self.__class__.__name__

    error = self.evalError(prediction,groundTruth,self)
    popped = self.errorStore.pushToEnd({"error":error})
    if not popped is None:
      #Subtract error that dropped out of the buffer
      self.error -= popped["error"]
    self.error+= error
    self.averageError =  float(self.error)/self.errorStore.elements
    self.steps+=1
    
    return self.averageError



class MetricMovingMode(AggregateMetric):
  """
      computes error metric based on moving mode prediction

  """


  def __init__(self, metricSpec):

    super(MetricMovingMode, self).__init__(metricSpec)

    self.mode_window = 100
    if metricSpec.params.has_key('mode_window'):
      assert metricSpec.params['mode_window'] >= 1
      self.mode_window = metricSpec.params['mode_window']
      
    # Only supports one stepsize
    assert len(self._predictionSteps) == 1

    # Construct moving average instance
    self._movingMode = _MovingMode(self.mode_window)

  def getMetric(self):
    return self._subErrorMetrics[0].getMetric()

  def addInstance(self, groundTruth, prediction, record = None, result = None):

    # If missing data,
    if groundTruth == SENTINEL_VALUE_FOR_MISSING_DATA:
      return self._subErrorMetrics[0].aggregateError

    if self.verbosity > 0:
      print "groundTruth:\n%s\nPredictions:\n%s\n%s\n" % (groundTruth, prediction,
                                                          self.getMetric())

    # Use ground truth from 'steps' steps ago as our most recent ground truth
    lastGT = self._getShiftedGroundTruth(groundTruth)
    if lastGT is None:
      return self._subErrorMetrics[0].aggregateError

    mode = self._movingMode(lastGT)

    result = self._subErrorMetrics[0].addInstance(groundTruth, mode, record)
    return result



class MetricTrivial(AggregateMetric):
  """
  computes a metric against the ground truth N steps ago. The metric to
  compute is designated by the 'errorMetric' entry in the metric params.
  """

  def __init__(self, metricSpec):

    # This metric assumes a default 'steps' of 1
    if not 'steps' in metricSpec.params:
      metricSpec.params['steps'] = 1

    super(MetricTrivial, self).__init__(metricSpec)

    # Only supports one stepsize
    assert len(self._predictionSteps) == 1

    # Must have a suberror metric
    assert self._subErrorMetrics is not None, "This metric requires that you" \
        + " specify the name of another base metric  via the 'errorMetric' " \
        + " parameter."

  def getMetric(self):
    return self._subErrorMetrics[0].getMetric()

  def addInstance(self, groundTruth, prediction, record = None, result = None):

    # Use ground truth from 'steps' steps ago as our "prediction"
    prediction = self._getShiftedGroundTruth(groundTruth)

    if self.verbosity > 0:
      print "groundTruth:\n%s\nPredictions:\n%s\n%s\n" % (groundTruth,
                                            prediction, self.getMetric())
    # If missing data,
    if groundTruth == SENTINEL_VALUE_FOR_MISSING_DATA:
      return self._subErrorMetrics[0].aggregateError

    # Our "prediction" is simply what happened 'steps' steps ago
    return self._subErrorMetrics[0].addInstance(groundTruth, prediction, record)



class MetricTwoGram(AggregateMetric):
  """
  computes error metric based on one-grams. The groundTruth passed into
  this metric is the encoded output of the field (an array of 1's and 0's).

  """


  def __init__(self, metricSpec):

    # This metric assumes a default 'steps' of 1
    if not 'steps' in metricSpec.params:
      metricSpec.params['steps'] = 1

    super(MetricTwoGram, self).__init__(metricSpec)
    
    # Only supports 1 stepsize
    assert len(self._predictionSteps) == 1

    # Must supply the predictionField
    assert(metricSpec.params.has_key('predictionField'))
    self.predictionField = metricSpec.params['predictionField']
    self.twoGramDict = dict()

  def getMetric(self):
    return self._subErrorMetrics[0].getMetric()

  def addInstance(self, groundTruth, prediction, record = None, result = None):

    # If missing data return previous error (assuming one gram will always
    #  receive an instance of ndarray)
    if groundTruth.any() == False:
      return self._subErrorMetrics[0].aggregateError

    # Get actual ground Truth value from record. For this metric, the
    # "groundTruth" parameter is the encoder output and we use actualGroundTruth
    # to hold the input to the encoder (either a scalar or a category string).
    #
    # We will use 'groundTruthKey' (the stringified encoded value of
    # groundTruth) as the key for our one-gram dict and the 'actualGroundTruth'
    # as the values in our dict, which are used to compute our prediction.
    actualGroundTruth = record[self.predictionField]

    # convert binary array to a string
    groundTruthKey = str(groundTruth)

    # Get the ground truth key from N steps ago, that is what we will base
    #  our prediction on. Note that our "prediction" is the prediction for the
    #  current time step, to be compared to actualGroundTruth
    prevGTKey = self._getShiftedGroundTruth(groundTruthKey)

    # -------------------------------------------------------------------------
    # Get the prediction based on the previously known ground truth
    # If no previous, just default to "" or 0, depending on the groundTruth
    #  data type.
    if prevGTKey == None:
      if isinstance(actualGroundTruth,str):
        pred = ""
      else:
        pred = 0

    # If the previous was never seen before, create a new dict for it.
    elif not prevGTKey in self.twoGramDict:
      if isinstance(actualGroundTruth,str):
        pred = ""
      else:
        pred = 0
      # Create a new dict for it
      self.twoGramDict[prevGTKey] = {actualGroundTruth:1}

    # If it was seen before, compute the prediction from the past history
    else:
      # Find most often occurring 1-gram
      if isinstance(actualGroundTruth,str):
        # Get the most frequent category that followed the previous timestep
        twoGramMax = max(self.twoGramDict[prevGTKey].items(), key=itemgetter(1))
        pred = twoGramMax[0]

      else:
        # Get average of all possible values that followed the previous
        # timestep
        pred = sum(self.twoGramDict[prevGTKey].iterkeys())
        pred /= len(self.twoGramDict[prevGTKey])

      # Add current ground truth to dict
      if actualGroundTruth in self.twoGramDict[prevGTKey]:
        self.twoGramDict[prevGTKey][actualGroundTruth] += 1
      else:
        self.twoGramDict[prevGTKey][actualGroundTruth] = 1

    if self.verbosity > 0:
      print "\nencoding:%s\nactual:%s\nprevEncoding:%s\nprediction:%s\nmetric:%s" % \
          (groundTruth, actualGroundTruth, prevGTKey, pred, self.getMetric())

    return self._subErrorMetrics[0].addInstance(actualGroundTruth, pred, record)



class MetricAccuracy(AggregateMetric):
  """
  computes simple accuracy for an enumerated type. all inputs are treated as
  discrete members of a set, therefore for example 0.5 is only a correct
  response if the ground truth is exactly 0.5. Inputs can be strings, integers,
  or reals
  """

  def accumulate(self, groundTruth, prediction, accumulatedError, historyBuffer, result = None):

    # This is really an accuracy measure rather than an "error" measure
    error = 1.0 if groundTruth == prediction else 0.0
    accumulatedError += error

    if historyBuffer is not None:
      historyBuffer.append(error)
      if len(historyBuffer) > self.spec.params["window"] :
        accumulatedError -= historyBuffer.popleft()

    return accumulatedError

  def aggregate(self, accumulatedError, historyBuffer, steps):
    n = steps
    if historyBuffer is not None:
      n = len(historyBuffer)

    return accumulatedError/ float(n)



class MetricAveError(AggregateMetric):
  """Simply the inverse of the Accuracy metric
        More consistent with scalar metrics because
        they all report an error to be minimized"""

  def accumulate(self, groundTruth, prediction, accumulatedError, historyBuffer, result = None):

    error = 1.0 if groundTruth != prediction else 0.0
    accumulatedError += error

    if historyBuffer is not None:
      historyBuffer.append(error)
      if len(historyBuffer) > self.spec.params["window"] :
        accumulatedError -= historyBuffer.popleft()

    return accumulatedError

  def aggregate(self, accumulatedError, historyBuffer, steps):
    n = steps
    if historyBuffer is not None:
      n = len(historyBuffer)

    return accumulatedError/ float(n)



class MetricNegAUC(AggregateMetric):
  """ Computes -1 * AUC (Area Under the Curve) of the ROC (Receiver Operator
      Characteristics) curve. We compute -1 * AUC because metrics are optimized
      to be LOWER when running hypersearch.

      For this, we assuming that category 1 is the "positive" category and
      we are generating an ROC curve with the TPR (True Positive Rate) of
      category 1 on the y-axis and the FPR (False Positive Rate) on the x-axis.
  """

  def accumulate(self, groundTruth, prediction, accumulatedError, historyBuffer, result = None):
    """ Accumulate history of groundTruth and "prediction" values.

    For this metric, groundTruth is the actual category and "prediction" is a
    dict containing one top-level item with a key of 0 (meaning this is the
    0-step classificaton) and a value which is another dict, which contains the
    probability for each category as output from the classifier. For example,
    this is what "prediction" would be if the classifier said that category 0
    had a 0.6 probability and category 1 had a 0.4 probability: {0:0.6, 1: 0.4}
    """

    # We disable it within aggregate() if we find that the classifier classes
    #  are not compatible with AUC calculations.
    if self.disabled:
      return 0

    # Just store the groundTruth, probability into our history buffer. We will
    #  wait until aggregate gets called to actually compute AUC.
    if historyBuffer is not None:
      historyBuffer.append((groundTruth, prediction[0]))
      if len(historyBuffer) > self.spec.params["window"] :
        historyBuffer.popleft()

    # accumulatedError not used in this metric
    return 0

  def aggregate(self, accumulatedError, historyBuffer, steps):

    # If disabled, do nothing.
    if self.disabled:
      return 0.0

    if historyBuffer is not None:
      n = len(historyBuffer)
    else:
      return 0.0

    # For performance reasons, only re-compute this every 'computeEvery' steps
    frequency = self.spec.params.get('computeEvery', 1)
    if ((steps+1) % frequency) != 0:
      return self.aggregateError

    # Compute the ROC curve and the area underneath it
    actuals = [gt for (gt, probs) in historyBuffer]
    classes = np.unique(actuals)

    # We can only compute ROC when we have at least 1 sample of each category
    if len(classes) < 2:
      return -1 * 0.5

    # Print warning the first time this metric is asked to be computed on a
    #  problem with more than 2 classes
    if sorted(classes) != [0,1]:
      print "WARNING: AUC only implemented for binary classifications where " \
          "the categories are category 0 and 1. In this network, the " \
          "categories are: %s" % (classes)
      print "WARNING: Computation of this metric is disabled for the remainder of " \
            "this experiment."
      self.disabled = True
      return 0.0

    # Compute the ROC and AUC. Note that because we are online, there's a
    #  chance that some of the earlier classification probabilities don't
    #  have the True class (category 1) yet because it hasn't been seen yet.
    #  Therefore, we use probs.get() with a default value of 0.
    scores = [probs.get(1, 0) for (gt, probs) in historyBuffer]
    (fpr, tpr, thresholds) = roc.ROCCurve(actuals, scores)
    auc = roc.AreaUnderCurve(fpr, tpr)

    # Debug?
    if False:
      print
      print "AUC metric debug info (%d steps):" % (steps)
      print " actuals:", actuals
      print " probabilities:", ["%.2f" % x for x in scores]
      print " fpr:", fpr
      print " tpr:", tpr
      print " thresholds:", thresholds
      print " AUC:", auc

    return -1 * auc



class MetricMultiStep(AggregateMetric):
  """
  This is an "uber" metric which is used to apply one of the other basic
  metrics to a specific step in a multi-step prediction.

  The specParams are expected to contain:
      'errorMetric': name of basic metric to apply
      'steps':       compare prediction['steps'] to the current
                      ground truth.

  Note that the metrics manager has already performed the time shifting
  for us - it passes us the prediction element from 'steps' steps ago
  and asks us to compare that to the current ground truth.
  
  When multiple steps of prediction are requested, we average the results of
  the underlying metric for each step. 
  """
  def __init__(self, metricSpec):

    super(MetricMultiStep, self).__init__(metricSpec)

    assert self._subErrorMetrics is not None

  def getMetric(self):
    return {'value': self.aggregateError, "stats" : {"steps" : self.steps}}


  def addInstance(self, groundTruth, prediction, record = None, result = None):
    
    # If missing data,
    if groundTruth == SENTINEL_VALUE_FOR_MISSING_DATA:
      return self.aggregateError

    # Get the prediction for this time step
    aggErrSum = 0
    try:
      for step, subErrorMetric in \
                  zip(self._predictionSteps, self._subErrorMetrics):
        
        stepPrediction = prediction[step]
        # Unless this is a custom_error_metric, when we have a dict of 
        #  probabilities, get the most probable one. For custom error metrics, 
        #  we pass the probabilities in so that it can decide how best to deal with
        #  them.
        if isinstance(stepPrediction, dict) \
            and not isinstance(subErrorMetric, CustomErrorMetric):
          predictions = [(prob,value) for (value, prob) in \
                                                    stepPrediction.iteritems()]
          predictions.sort()
          stepPrediction = predictions[-1][1]
          
        # Get sum of the errors
        aggErr = subErrorMetric.addInstance(groundTruth, stepPrediction, record, result)
        if self.verbosity >= 2:
          print "MetricMultiStep %s: aggErr for stepSize %d: %s" % \
                  (self._predictionSteps, step, aggErr)
        
        aggErrSum += aggErr
    except:
      pass
  

    # Return average aggregate error across all step sizes
    self.aggregateError = aggErrSum / len(self._subErrorMetrics)
    if self.verbosity >= 2:
      print "MetricMultiStep %s: aggErrAvg: %s" % (self._predictionSteps, 
                                                   self.aggregateError)
    self.steps += 1

    if self.verbosity >= 1:
      print "\nMetricMultiStep %s: \n  groundTruth:  %s\n  Predictions:  %s" \
            "\n  Metric:  %s" % (self._predictionSteps, groundTruth, prediction, 
                                 self.getMetric())

    return self.aggregateError



class MetricMultiStepProbability(AggregateMetric):
  """
  This is an "uber" metric which is used to apply one of the other basic
  metrics to a specific step in a multi-step prediction.

  The specParams are expected to contain:
      'errorMetric': name of basic metric to apply
      'steps':       compare prediction['steps'] to the current
                      ground truth.

  Note that the metrics manager has already performed the time shifting
  for us - it passes us the prediction element from 'steps' steps ago
  and asks us to compare that to the current ground truth.
  """
  def __init__(self, metricSpec):


    # Default window should be 1
    if not 'window' in metricSpec.params:
      metricSpec.params['window'] = 1

    super(MetricMultiStepProbability, self).__init__(metricSpec)

    # Must have a suberror metric
    assert self._subErrorMetrics is not None, "This metric requires that you" \
        + " specify the name of another base metric  via the 'errorMetric' " \
        + " parameter."

    # Force all subErrorMetric windows to 1. This is necessary because by 
    #  default they each do their own history averaging assuming that their
    #  addInstance() gets called once per interation. But, in this metric
    #  we actually call into each subErrorMetric multiple times per iteration
    for subErrorMetric in self._subErrorMetrics:
      subErrorMetric.window = 1
      subErrorMetric.spec.params['window'] = 1

    self._movingAverage = MovingAverage(self.window)

  def getMetric(self):
    return {'value': self.aggregateError, "stats" :
            {"steps" : self.steps}}

  def addInstance(self, groundTruth, prediction, record = None, result = None):

    # If missing data,
    if groundTruth == SENTINEL_VALUE_FOR_MISSING_DATA:
      return self.aggregateError

    if self.verbosity >= 1:
      print "\nMetricMultiStepProbability %s: \n  groundTruth:  %s\n  " \
            "Predictions:  %s" % (self._predictionSteps, groundTruth, 
                                  prediction)
      
    # Get the aggregateErrors for all requested step sizes and average them
    aggErrSum = 0
    for step, subErrorMetric in \
                zip(self._predictionSteps, self._subErrorMetrics):
      
      stepPrediction = prediction[step]
  
      # If it's a dict of probabilities, get the expected value
      error = 0
      if isinstance(stepPrediction, dict):
        expectedValue = 0
        # For every possible prediction multiply its error by its probability
        for (pred, prob) in stepPrediction.iteritems():
          error += subErrorMetric.addInstance(groundTruth, pred, record) \
                    * prob
      else:
        error += subErrorMetric.addInstance(groundTruth, stepPrediction,
                                            record)
  
      if self.verbosity >= 2:
          print ("MetricMultiStepProbability %s: aggErr for stepSize %d: %s" % 
                 (self._predictionSteps, step, error))

      aggErrSum += error


    # Return aggregate error
    avgAggErr = aggErrSum / len(self._subErrorMetrics)
    self.aggregateError = self._movingAverage(avgAggErr)
    if self.verbosity >= 2:
      print ("MetricMultiStepProbability %s: aggErr over all steps, this "
             "iteration (%d): %s" % (self._predictionSteps, self.steps, avgAggErr))
      print ("MetricMultiStepProbability %s: aggErr moving avg: %s" % 
             (self._predictionSteps, self.aggregateError))
    self.steps += 1
    
    if self.verbosity >= 1:
      print "MetricMultiStepProbability %s: \n  Error: %s\n  Metric:  %s" % \
              (self._predictionSteps, avgAggErr, self.getMetric())

    return self.aggregateError



class MetricMulti(MetricsIface):
  """Multi metric can combine multiple other (sub)metrics and 
     weight them to provide combined score."""

  def __init__(self, metricSpec):
    """MetricMulti constructor using metricSpec is not allowed."""
    raise ValueError("MetricMulti cannot be constructed from metricSpec string! "
                     "Use MetricMulti(weights,metrics) constructor instead.")

  def __init__(self, weights, metrics, window=None):
    """MetricMulti 
       @param weights - [list of floats] used as weights
       @param metrics - [list of submetrics] 
       @param window - (opt) window size for moving average, or None when disabled
    """
    if (weights is None or not isinstance(weights, list) or 
                          not len(weights) > 0 or
                          not isinstance(weights[0], float)):
      raise ValueError("MetricMulti requires 'weights' parameter as a [list of floats]")
    self.weights = weights

    if (metrics is None or not isinstance(metrics, list) or
                          not len(metrics) > 0 or
                          not isinstance(metrics[0], MetricsIface)):
      raise ValueError("MetricMulti requires 'metrics' parameter as a [list of Metrics]")
    self.metrics = metrics
    if window is not None:
      self.movingAvg = MovingAverage(windowSize=window)
    else:
      self.movingAvg = None


  def addInstance(self, groundTruth, prediction, record = None, result = None):
    err = 0.0
    subResults = [m.addInstance(groundTruth, prediction, record) for m in self.metrics]
    for i in xrange(len(self.weights)):
      if subResults[i] is not None:
        err += subResults[i]*self.weights[i]
      else: # submetric returned None, propagate
        self.err = None
        return None

    if self.verbosity > 2:
      print "IN=",groundTruth," pred=",prediction,": w=",self.weights[i]," metric=",self.metrics[i]," value=",m," err=",err
    if self.movingAvg is not None:
      err=self.movingAvg(err)
    self.err = err
    return err


  def __repr__(self):
    return "MetricMulti(weights=%s, metrics=%s)" % (self.weights, self.metrics) 


  def getMetric(self):
    return {'value': self.err, "stats" : {"weights" : self.weights}}




# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2015, Numenta, Inc.  Unless you have purchased from
# Numenta, Inc. a separate commercial license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

"""Region for computing the anomaly score."""

import numpy

from nupic.algorithms import anomaly
from nupic.bindings.regions.PyRegion import PyRegion



class AnomalyRegion(PyRegion):
  """Region for computing the anomaly score."""


  @classmethod
  def getSpec(cls):
    return {
        "description": ("Region that computes anomaly scores from temporal "
                        "memory."),
        "singleNodeOnly": True,
        "inputs": {
            "activeColumns": {
                "description": "The currently active columns.",
                "regionLevel": True,
                "dataType": "Real32",
                "count": 0,
                "required": True,
                "isDefaultInput": False,
                "requireSplitterMap": False,
            },
            "predictedColumns": {
                "description": "The currently predicted columns.",
                "regionLevel": True,
                "dataType": "Real32",
                "count": 0,
                "required": True,
                "isDefaultInput": False,
                "requireSplitterMap": False,
            },
        },
        "outputs": {
            "rawAnomalyScore": {
                "description": "The raw anomaly score.",
                "dataType": "Real32",
                "count": 1,
                "regionLevel": True,
                "isDefaultOutput": True,
            },
        },
        "parameters": {
        },
        "commands": {
        },
    }


  def __init__(self, *args, **kwargs):
    self.prevPredictedColumns = numpy.zeros([], dtype="float32")


  def __eq__(self, other):
    for k, v1 in self.__dict__.iteritems():
      if not k in other.__dict__:
        return False
      v2 = getattr(other, k)
      if isinstance(v1, numpy.ndarray):
        if v1.dtype != v2.dtype:
          return False
        if not numpy.isclose(v1, v2).all():
          return False
      else:
        if type(v1) != type(v2):
          return False
        if v1 != v2:
          return False
    return True


  def __ne__(self, other):
    return not self == other


  @classmethod
  def read(cls, proto):
    anomalyRegion = object.__new__(cls)
    anomalyRegion.prevPredictedColumns = numpy.array(proto.prevPredictedColumns)
    return anomalyRegion


  def write(self, proto):
    proto.prevPredictedColumns = self.prevPredictedColumns.tolist()


  def initialize(self, inputs, outputs):
    pass


  def compute(self, inputs, outputs):
    activeColumns = inputs["activeColumns"].nonzero()[0]

    rawAnomalyScore = anomaly.computeRawAnomalyScore(
        activeColumns, self.prevPredictedColumns)

    self.prevPredictedColumns = inputs["predictedColumns"].nonzero()[0]

    outputs["rawAnomalyScore"][0] = rawAnomalyScore



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

# This script implements PredictionMetricsManager, a helper class that handles
# pooling of multiple record and field prediction metrics calculators

import logging
import copy
import pprint
from collections import (namedtuple,
                         deque)

from nupic.data.inference_shifter import InferenceShifter
from nupic.frameworks.opf import metrics

from opfutils import InferenceType, InferenceElement



# MetricValueElement class
#
# Represents an individual metric value element in a list returned by
# PredictionMetricsManager.getMetrics()
#
# spec:           A MetricSpec value (a copy) that was used to construct
#                 the metric instance that generated the metric value
# value:          The metric value
MetricValueElement = namedtuple("MetricValueElement", ["spec", "value"])



class MetricsManager(object):
  """ This is a class to handle the computation of metrics properly. This class
  takes in an inferenceType, and it assumes that it is associcated with a single
  model """

  # Map from inference element to sensor input element. This helps us find the
  # appropriate ground truth field for a given inference element

  def __init__(self, metricSpecs, fieldInfo, inferenceType):
    """
    Constructs a Metrics Manager

    Parameters:
    -----------------------------------------------------------------------
    metricSpecs:    A sequence of MetricSpecs that specify which metrics should
                    be calculated

    inferenceType:  An opfutils.inferenceType value that specifies the inference
                    type of the associated model. This affects how metrics are
                    calculated. FOR EXAMPLE, temporal models save the inference
                    from the previous timestep to match it to the ground truth
                    value in the current timestep
    """

    self.__metricSpecs = []
    self.__metrics = []
    self.__metricLabels = []

    # Maps field names to indices. Useful for looking up input/predictions by
    # field name
    self.__fieldNameIndexMap = dict( [(info.name, i) \
                                      for i, info in enumerate(fieldInfo)] )

    self.__constructMetricsModules(metricSpecs)
    self.__currentGroundTruth = None
    self.__currentInference = None
    self.__currentResult = None

    self.__isTemporal = InferenceType.isTemporal(inferenceType)
    if self.__isTemporal:
      self.__inferenceShifter = InferenceShifter()


  def update(self, results):
    """
    Compute the new metrics values, given the next inference/ground-truth values

    Parameters:
    -----------------------------------------------------------------------
    results:  An opfutils.ModelResult object that was computed during the last
              iteration of the model

    Returns:  A dictionary where each key is the metric-name, and the values are
              it scalar value.

    """

    #print "\n\n---------------------------------------------------------------"
    #print "Model results: \nrawInput:%s \ninferences:%s" % \
    #      (pprint.pformat(results.rawInput), pprint.pformat(results.inferences))
          
    self._addResults(results)

    if  not self.__metricSpecs \
        or self.__currentInference is None:
      return {}

    metricResults = {}
    for metric, spec, label in zip(self.__metrics,
                                   self.__metricSpecs,
                                   self.__metricLabels):

      inferenceElement = spec.inferenceElement
      field = spec.field
      groundTruth = self._getGroundTruth(inferenceElement)
      inference = self._getInference(inferenceElement)
      rawRecord = self._getRawGroundTruth()
      result = self.__currentResult
      if field:
        if type(inference) in (list, tuple):
          if field in self.__fieldNameIndexMap:
            # NOTE: If the predicted field is not fed in at the bottom, we
            #  won't have it in our fieldNameIndexMap
            fieldIndex = self.__fieldNameIndexMap[field]
            inference = inference[fieldIndex]
          else:
            inference = None
        if groundTruth is not None:
          if type(groundTruth) in (list, tuple):
            if field in self.__fieldNameIndexMap:
              # NOTE: If the predicted field is not fed in at the bottom, we
              #  won't have it in our fieldNameIndexMap
              fieldIndex = self.__fieldNameIndexMap[field]
              groundTruth = groundTruth[fieldIndex]
            else:
              groundTruth = None
          else:
            # groundTruth could be a dict based off of field names
            groundTruth = groundTruth[field]

      metric.addInstance(groundTruth=groundTruth,
                         prediction=inference,
                         record=rawRecord,
                         result=result)

      metricResults[label] = metric.getMetric()['value']

    return metricResults


  def getMetrics(self):
    """ Gets the current metric values

    Returns: A dictionary where each key is the metric-name, and the values are
              it scalar value. Same as the output of update()
    """

    result = {}

    for metricObj, label in zip(self.__metrics, self.__metricLabels):
      value = metricObj.getMetric()
      result[label] = value['value']

    return result


  def getMetricDetails(self, metricLabel):
    """ Gets detailed info about a given metric, in addition to its value. This
    may including any statistics or auxilary data that are computed for a given
    metric

    Parameters:
    -----------------------------------------------------------------------
    metricLabel:   The string label of the given metric (see metrics.MetricSpec)

    Returns:  A dictionary of metric information, as returned by
              opf.metric.Metric.getMetric()
    """
    try:
      metricIndex = self.__metricLabels.index(metricLabel)
    except IndexError:
      return None

    return self.__metrics[metricIndex].getMetric()


  def getMetricLabels(self):
    """ Return the list of labels for the metrics that are being calculated"""
    return tuple(self.__metricLabels)


  def _addResults(self, results):
    """
    Stores the current model results in the manager's internal store

    Parameters:
    -----------------------------------------------------------------------
    results:  A ModelResults object that contains the current timestep's
              input/inferences
    """
    # -----------------------------------------------------------------------
    # If the model potentially has temporal inferences.
    if self.__isTemporal:
      shiftedInferences = self.__inferenceShifter.shift(results).inferences
      self.__currentResult = copy.deepcopy(results)
      self.__currentResult.inferences = shiftedInferences
      self.__currentInference = shiftedInferences

    # -----------------------------------------------------------------------
    # The current model has no temporal inferences.
    else:
      self.__currentResult = copy.deepcopy(results)
      self.__currentInference = copy.deepcopy(results.inferences)

    # -----------------------------------------------------------------------
    # Save the current ground-truth results
    self.__currentGroundTruth = copy.deepcopy(results)


  def _getGroundTruth(self, inferenceElement):
    """
    Get the actual value for this field

    Parameters:
    -----------------------------------------------------------------------
    sensorInputElement:       The inference element (part of the inference) that
                            is being used for this metric
    """
    sensorInputElement = InferenceElement.getInputElement(inferenceElement)
    if sensorInputElement is None:
      return None
    return getattr(self.__currentGroundTruth.sensorInput, sensorInputElement)


  def _getInference(self, inferenceElement):
    """
    Get what the inferred value for this field was

    Parameters:
    -----------------------------------------------------------------------
    inferenceElement:       The inference element (part of the inference) that
                            is being used for this metric
    """
    if self.__currentInference is not None:
      return self.__currentInference.get(inferenceElement, None)

    return None


  def _getRawGroundTruth(self):
    """
    Get what the inferred value for this field was

    Parameters:
    -----------------------------------------------------------------------
    inferenceElement:       The inference element (part of the inference) that
                            is being used for this metric
    """

    return self.__currentGroundTruth.rawInput


  def __constructMetricsModules(self, metricSpecs):
    """
    Creates the required metrics modules

    Parameters:
    -----------------------------------------------------------------------
    metricSpecs:
      A sequence of MetricSpec objects that specify which metric modules to
      instantiate
    """
    if not metricSpecs:
      return

    self.__metricSpecs = metricSpecs
    for spec in metricSpecs:
      if not InferenceElement.validate(spec.inferenceElement):
        raise ValueError("Invalid inference element for metric spec: %r" %spec)

      self.__metrics.append(metrics.getModule(spec))
      self.__metricLabels.append(spec.getLabel())



def test():

  _testMetricsMgr()
  _testTemporalShift()
  _testMetricLabels()

  return



def _testMetricsMgr():
  print "*Testing Metrics Managers*..."
  from nupic.data.fieldmeta import (
    FieldMetaInfo,
    FieldMetaType,
    FieldMetaSpecial)

  from nupic.frameworks.opf.metrics import MetricSpec
  from nupic.frameworks.opf.opfutils import ModelResult, SensorInput
  onlineMetrics = (MetricSpec(metric="aae", inferenceElement='', \
                              field="consumption", params={}),)

  print "TESTING METRICS MANAGER (BASIC PLUMBING TEST)..."

  modelFieldMetaInfo = (
    FieldMetaInfo(name='temperature',
                  type=FieldMetaType.float,
                  special=FieldMetaSpecial.none),
    FieldMetaInfo(name='consumption',
              type=FieldMetaType.float,
              special=FieldMetaSpecial.none)
  )

  # -----------------------------------------------------------------------
  # Test to make sure that invalid InferenceElements are caught
  try:
    MetricsManager(
    metricSpecs=onlineMetrics,
    fieldInfo=modelFieldMetaInfo,
    inferenceType=InferenceType.TemporalNextStep)
  except ValueError:
    print "Caught bad inference element: PASS"


  print
  onlineMetrics = (MetricSpec(metric="aae",
                              inferenceElement=InferenceElement.prediction,
                              field="consumption", params={}),)

  temporalMetrics = MetricsManager(
    metricSpecs=onlineMetrics,
    fieldInfo=modelFieldMetaInfo,
    inferenceType=InferenceType.TemporalNextStep)



  inputs = [
    {
      'groundTruthRow' : [9, 7],

      'predictionsDict' : {
        InferenceType.TemporalNextStep: [12, 17]
      }
    },

    {
      'groundTruthRow' : [12, 17],

      'predictionsDict' : {
        InferenceType.TemporalNextStep: [14, 19]
      }
    },

    {
      'groundTruthRow' : [14, 20],

      'predictionsDict' : {
        InferenceType.TemporalNextStep: [16, 21]
      }
    },

    {
      'groundTruthRow' : [9, 7],

      'predictionsDict' : {
        InferenceType.TemporalNextStep:None
      }
    },
  ]


  for element in inputs:
    groundTruthRow=element['groundTruthRow']
    tPredictionRow=element['predictionsDict'][InferenceType.TemporalNextStep]

    result = ModelResult(sensorInput=SensorInput(dataRow=groundTruthRow,
                                                 dataEncodings=None,
                                                 sequenceReset=0,
                                                 category=None),
                         inferences={'prediction':tPredictionRow})

    temporalMetrics.update(result)

  assert temporalMetrics.getMetrics().values()[0] == 15.0 / 3.0, \
          "Expected %f, got %f" %(15.0/3.0,
                                  temporalMetrics.getMetrics().values()[0])
  print "ok"

  return



def _testTemporalShift():
  """ Test to see if the metrics manager correctly shifts records for multistep
  prediction cases
  """
  print "*Testing Multistep temporal shift*..."
  from nupic.data.fieldmeta import (
    FieldMetaInfo,
    FieldMetaType,
    FieldMetaSpecial)

  from nupic.frameworks.opf.metrics import MetricSpec
  from nupic.frameworks.opf.opfutils import ModelResult, SensorInput
  onlineMetrics = ()

  modelFieldMetaInfo = (
    FieldMetaInfo(name='consumption',
              type=FieldMetaType.float,
              special=FieldMetaSpecial.none),)

  mgr = MetricsManager(metricSpecs=onlineMetrics,
                       fieldInfo=modelFieldMetaInfo,
                       inferenceType=InferenceType.TemporalMultiStep)

  groundTruths = [{'consumption':i} for i in range(10)]
  oneStepInfs = reversed(range(10))
  threeStepInfs = range(5, 15)

  for iterNum, gt, os, ts in zip(xrange(10), groundTruths,
                              oneStepInfs, threeStepInfs):
    inferences = {InferenceElement.multiStepPredictions:{1: os, 3: ts}}
    sensorInput = SensorInput(dataDict = [gt])
    result = ModelResult(sensorInput=sensorInput, inferences=inferences)
    mgr.update(result)

    assert mgr._getGroundTruth(InferenceElement.multiStepPredictions)[0] == gt
    if iterNum < 1:
      #assert mgr._getInference(InferenceElement.multiStepPredictions) is None
      assert mgr._getInference(InferenceElement.multiStepPredictions)[1] is None
    else:
      prediction = mgr._getInference(InferenceElement.multiStepPredictions)[1]
      assert prediction == 10 - iterNum

    if iterNum < 3:
      inference = mgr._getInference(InferenceElement.multiStepPredictions)
      assert inference is None or inference[3] is None
    else:
      prediction = mgr._getInference(InferenceElement.multiStepPredictions)[3]
      assert prediction == iterNum + 2



def _testMetricLabels():
  print "\n*Testing Metric Label Generation*..."

  from nupic.frameworks.opf.metrics import MetricSpec

  testTuples = [
    (MetricSpec('rmse', InferenceElement.prediction, 'consumption'),
     "prediction:rmse:field=consumption"),
    (MetricSpec('rmse', InferenceElement.classification),
     "classification:rmse"),
    (MetricSpec('rmse', InferenceElement.encodings, 'pounds',
                params=dict(window=100)),
     "encodings:rmse:window=100:field=pounds"),
    (MetricSpec('aae', InferenceElement.prediction, 'pounds',
                params=dict(window=100, paramA = 10.2, paramB = 20)),
     "prediction:aae:paramA=10.2:paramB=20:window=100:field=pounds"),
    (MetricSpec('aae', InferenceElement.prediction,'pounds',
                params={'window':100, 'paramA':10.2, '1paramB':20}),
     "prediction:aae:1paramB=20:paramA=10.2:window=100:field=pounds"),
    (MetricSpec('aae', InferenceElement.prediction,'pounds',
                params=dict(window=100, paramA = 10.2, paramB =-20)),
     "prediction:aae:paramA=10.2:paramB=-20:window=100:field=pounds"),
    (MetricSpec('aae', InferenceElement.prediction, 'pounds',
                params=dict(window=100, paramA = 10.2, paramB ='square')),
     "prediction:aae:paramA=10.2:paramB='square':window=100:field=pounds"),
  ]

  for test in testTuples:
    try:
      assert test[0].getLabel() == test[1]
    except:
      print "Failed Creating label"
      print "Expected %s \t Got %s" % (test[1], test[0].getLabel())
      return

  print "ok"



if __name__ == "__main__":
  test()



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

raise Exception("XERROR -- removing for NuPIC 2")


import numpy
from numpy.linalg import svd

log2pi = numpy.log(2.0 * numpy.pi)

def cov(data, mean=None, prior=None, dfOffset=-1):
  x = numpy.asmatrix(data)
  if mean is None: mean = x.mean()
  xc = x - mean
  xxt = xc.T * xc
  n = x.shape[0]
  if prior is not None:
    assert len(prior) == 2, "Must be of the form (n, SS)"
    n += prior[0]
    print prior[1].shape
    xxt += prior[1]
  n += dfOffset
  assert n > 0
  return (1.0 / n) * xxt

def getRank(d, s):
  if numpy.min(s) > 0: return d
  else: return max(numpy.argmin(s > 0), 1)

class mvn(object):
  def __init__(self, mean, varcov):
    if mean is not None: self.mean = numpy.asarray(mean)
    else: self.mean = None
    varcov = numpy.asmatrix(varcov)
    self.d = varcov.shape[1]
    self.varcov = varcov
    self.u, self.s, self.vt = svd(varcov, full_matrices=0, compute_uv=1)
    self.rank = getRank(self.d, self.s)

  def __str__(self):
    return "Mean:\n" + str(self.mean) + "\nCovariance:\n" + str(self.varcov)

  def __repr__(self):
    return "Mean:\n" + repr(self.mean) + "\nCovariance:\n" + repr(self.varcov)

  def limitRank(self, minvar):
    if numpy.min(self.s) > minvar: self.rank = self.d
    else:
      self.rank = max(numpy.argmin(self.s > minvar), 1)

  def setRank(self, rank):
    assert rank <= self.d
    assert rank >= 1
    self.rank = rank

  def s0(self):
    s = numpy.zeros(len(self.s))
    s[0:self.rank] = self.s[0:self.rank]
    s[self.rank:] = 0
    return s

  def si(self):
    si = numpy.zeros(len(self.s))
    si[0:self.rank] = 1.0 / self.s[0:self.rank]
    si[self.rank:] = 0
    return si

  def sigma(self):
    return self.u * numpy.asmatrix(numpy.diag(self.s0())) * self.vt

  def sigmai(self):
    # return self.vt.T * numpy.asmatrix(numpy.diag(self.si())) * self.u.T
    return self.u * numpy.asmatrix(numpy.diag(self.si())) * self.vt

  def rightRoot(self):
    return numpy.asmatrix(numpy.diag(numpy.sqrt(self.s0()))) * self.vt

  def leftRoot(self):
    return self.u * numpy.asmatrix(numpy.diag(numpy.sqrt(self.s0())))

  def leftInvRoot(self):
    # return self.vt.T * numpy.asmatrix(numpy.diag(numpy.sqrt(self.si())))
    return self.u * numpy.asmatrix(numpy.diag(numpy.sqrt(self.si())))

  def rightInvRoot(self):
    # return numpy.asmatrix(numpy.diag(numpy.sqrt(self.si()))) * self.u.T
    return numpy.asmatrix(numpy.diag(numpy.sqrt(self.si()))) * self.vt

  def sample(self, r=None, n=1):
    if r is None: r = numpy.random
    z = r.normal(0, 1, (n, self.d))
    return z * self.rightRoot() + self.mean

  def center(self, x):
    x = numpy.asmatrix(x)
    assert x.shape[1] == self.d
    if self.mean is not None: return (x - self.mean)
    else: return x

  def whiten(self, x):
    xc = self.center(x)
    # Whiten.
    z = xc * self.leftInvRoot()
    return z

  def z2(self, x):
    z = self.whiten(x)
    # Avoid matrix multiplication, just square the rows.
    z = numpy.asarray(z)
    z2 = z * z
    return numpy.sum(z2, axis=1)

  def logDetSigma(self):
    return numpy.sum(numpy.log(self.s[0:self.rank]))

  def logDetInvRoot(self):
    return -0.5 * self.logDetSigma()

  def logK(self):
    return -0.5 * self.rank * log2pi

  def logLikelihood(self, x):
    z = numpy.asarray(self.whiten(x))
    # z2 = numpy.sum(z * z, axis=1)
    n = len(z)
    return -0.5 * ( n*(self.rank * log2pi + self.logDetSigma()) + numpy.sum(z*z) )

  def logLikelihoods(self, x):
    z2 = self.z2(x)
    return self.logK() + self.logDetInvRoot() - 0.5 * z2

  def logDensity(self, x):
    return self.logLikelihood(x)

class MaskBuilder(object):
  def __init__(self, d):
    self.indices = numpy.arange(d)

  def __getitem__(self, *args):
    return bits

class ConditionalMVN(object):
  def __init__(self, mvn):
    self.mvn = mvn

  class Impl(object):
    def __init__(self, mean1, mean2, Sigma11, Sigma12, Sigma22):
      Sigma11 = numpy.asmatrix(Sigma11)
      Sigma12 = numpy.asmatrix(Sigma12)
      Sigma22 = numpy.asmatrix(Sigma22)
      u22, s22, vt22 = svd(Sigma22, full_matrices=0, compute_uv=1)

      rank22 = getRank(Sigma22.shape[1], s22)
      s22i = numpy.zeros(len(s22))
      s22i[0:rank22] = 1.0 / s22[0:rank22]
      # Rest are zeroes.
      Sigma22i = u22 * numpy.asmatrix(numpy.diag(s22i)) * vt22

      self.mean1 = mean1
      self.mean2 = mean2
      self.Sigma11 = Sigma11
      self.Sigma12 = Sigma12
      self.Sigma22i = Sigma22i

    def getDistribution(self, given):
      given_from_mean = given - self.mean2

      # Keep means in row form.
      # mean = self.mean1 + self.Sigma12 * self.Sigma22i * given_from_mean
      mean = self.mean1 + given_from_mean * self.Sigma22i * self.Sigma12.transpose()
      varcov = self.Sigma11 - (self.Sigma12 * self.Sigma22i * (self.Sigma12.transpose()))

      return mvn(mean, varcov)

    def logDensity(self, x, given):
      return getDistribution(given).logDensity(x)

  def __getitem__(self, *args):
    d = self.mvn.d
    indices = numpy.arange(d).__getitem__(*args)
    bits = numpy.repeat(False, d)
    bits[indices] = True
    givenMask = bits # Should it be this way, or the other way around?
    varMask = ~givenMask
    C22 = self.mvn.varcov[givenMask, ...][..., givenMask]
    C12 = self.mvn.varcov[varMask, ...][..., givenMask]
    C11 = self.mvn.varcov[varMask, ...][..., varMask]
    return ConditionalMVN.Impl(self.mvn.mean[varMask], self.mvn.mean[givenMask],
      C11, C12, C22)



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

"""This script contains helper routines for testing algorithms"""

import time
import traceback

import numpy
from nupic.bindings.algorithms import SpatialPooler as CPPSpatialPooler
from nupic.bindings.math import GetNTAReal, Random as NupicRandom
from nupic.research.spatial_pooler import SpatialPooler as PySpatialPooler


realType = GetNTAReal()
uintType = "uint32"


def getNumpyRandomGenerator(seed = None):
  """
  Return a numpy random number generator with the given seed.
  If seed is None, set it randomly based on time. Regardless we log
  the actual seed and stack trace so that test failures are replicable.
  """
  if seed is None:
    seed = int((time.time()%10000)*10)
  print "Numpy seed set to:", seed, "called by",
  callStack = traceback.extract_stack(limit=3)
  print callStack[0][2], "line", callStack[0][1], "->", callStack[1][2]
  return numpy.random.RandomState(seed)


def convertPermanences(sourceSP, destSP):
  """
  Transfer the permanences from source to dest SP's. This is used in test
  routines to counteract some drift between implementations.
  We assume the two SP's have identical configurations/parameters.
  """
  numColumns = sourceSP.getNumColumns()
  numInputs = sourceSP.getNumInputs()
  for i in xrange(numColumns):
    potential = numpy.zeros(numInputs).astype(uintType)
    sourceSP.getPotential(i, potential)
    destSP.setPotential(i, potential)

    perm = numpy.zeros(numInputs).astype(realType)
    sourceSP.getPermanence(i, perm)
    destSP.setPermanence(i, perm)

  

def getSeed():
  """Generate and log a 32-bit compatible seed value."""
  seed = int((time.time()%10000)*10)
  print "New seed generated as:", seed, "called by",
  callStack = traceback.extract_stack(limit=3)
  print callStack[0][2], "line", callStack[0][1], "->", callStack[1][2]
  return seed

  
  
def convertSP(pySp, newSeed):
  """
  Given an instance of a python spatial_pooler return an instance of the CPP
  spatial_pooler with identical parameters.
  """
  columnDim = pySp._columnDimensions
  inputDim = pySp._inputDimensions
  numInputs = pySp.getNumInputs()
  numColumns = pySp.getNumColumns()
  cppSp = CPPSpatialPooler(inputDim, columnDim)
  cppSp.setPotentialRadius(pySp.getPotentialRadius())
  cppSp.setPotentialPct(pySp.getPotentialPct())
  cppSp.setGlobalInhibition(pySp.getGlobalInhibition())

  numActiveColumnsPerInhArea = pySp.getNumActiveColumnsPerInhArea()
  localAreaDensity = pySp.getLocalAreaDensity()
  if (numActiveColumnsPerInhArea > 0):
    cppSp.setNumActiveColumnsPerInhArea(numActiveColumnsPerInhArea)
  else:
    cppSp.setLocalAreaDensity(localAreaDensity)

  cppSp.setStimulusThreshold(pySp.getStimulusThreshold())
  cppSp.setInhibitionRadius(pySp.getInhibitionRadius())
  cppSp.setDutyCyclePeriod(pySp.getDutyCyclePeriod())
  cppSp.setMaxBoost(pySp.getMaxBoost())
  cppSp.setIterationNum(pySp.getIterationNum())
  cppSp.setIterationLearnNum(pySp.getIterationLearnNum())
  cppSp.setSpVerbosity(pySp.getSpVerbosity())
  cppSp.setUpdatePeriod(pySp.getUpdatePeriod())
  cppSp.setSynPermTrimThreshold(pySp.getSynPermTrimThreshold())
  cppSp.setSynPermActiveInc(pySp.getSynPermActiveInc())
  cppSp.setSynPermInactiveDec(pySp.getSynPermInactiveDec())
  cppSp.setSynPermBelowStimulusInc(pySp.getSynPermBelowStimulusInc())
  cppSp.setSynPermConnected(pySp.getSynPermConnected())
  cppSp.setMinPctOverlapDutyCycles(pySp.getMinPctOverlapDutyCycles())
  cppSp.setMinPctActiveDutyCycles(pySp.getMinPctActiveDutyCycles())

  boostFactors = numpy.zeros(numColumns).astype(realType)
  pySp.getBoostFactors(boostFactors)
  cppSp.setBoostFactors(boostFactors)

  overlapDuty = numpy.zeros(numColumns).astype(realType)
  pySp.getOverlapDutyCycles(overlapDuty)
  cppSp.setOverlapDutyCycles(overlapDuty)

  activeDuty = numpy.zeros(numColumns).astype(realType)
  pySp.getActiveDutyCycles(activeDuty)
  cppSp.setActiveDutyCycles(activeDuty)

  minOverlapDuty = numpy.zeros(numColumns).astype(realType)
  pySp.getMinOverlapDutyCycles(minOverlapDuty)
  cppSp.setMinOverlapDutyCycles(minOverlapDuty)

  minActiveDuty = numpy.zeros(numColumns).astype(realType)
  pySp.getMinActiveDutyCycles(minActiveDuty)
  cppSp.setMinActiveDutyCycles(minActiveDuty)

  for i in xrange(numColumns):
    potential = numpy.zeros(numInputs).astype(uintType)
    pySp.getPotential(i, potential)
    cppSp.setPotential(i, potential)

    perm = numpy.zeros(numInputs).astype(realType)
    pySp.getPermanence(i, perm)
    cppSp.setPermanence(i, perm)

  pySp._random = NupicRandom(newSeed)
  cppSp.seed_(newSeed)
  return cppSp



def CreateSP(imp, params):
  """
  Helper class for creating an instance of the appropriate spatial pooler using
  given parameters. 

  Parameters:
  ----------------------------
  imp:       Either 'py' or 'cpp' for creating the appropriate instance.
  params:    A dict for overriding constructor parameters. The keys must
             correspond to contructor parameter names.
  
  Returns the SP object.
  """
  if (imp == "py"):
    spClass = PySpatialPooler
  elif (imp == "cpp"):
    spClass = CPPSpatialPooler
  else:
    raise RuntimeError("unrecognized implementation")

  sp = spClass(**params)
  
  return sp




#! /usr/bin/env python
# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2015, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------
"""
A simple web server for interacting with NuPIC.
Note: Requires web.py to run (install using '$ pip install web.py')
"""
import os
import sys
# The following loop removes the nupic package from the
# PythonPath (sys.path). This is necessary in order to let web
# import the built in math module rather than defaulting to
# nupic.math
while True:
  try:
    sys.path.remove(os.path.dirname(os.path.realpath(__file__)))
  except:
    break

import datetime
import json
import web

from nupic.frameworks.opf.modelfactory import ModelFactory



g_models = {}



urls = (
    # Web UI
    "/models", "ModelHandler",
    r"/models/([-\w]*)", "ModelHandler",
    r"/models/([-\w]*)/run", "ModelRunner",
)



class ModelHandler(object):

  def GET(self):
    """
    /models

    returns:
    [model1, model2, model3, ...] list of model names
    """
    global g_models
    return json.dumps({"models": g_models.keys()})


  def POST(self, name):
    """
    /models/{name}

    schema:
    {
      "modelParams": dict containing model parameters
      "predictedFieldName": str
    }

    returns:
    {"success":name}
    """
    global g_models

    data = json.loads(web.data())
    modelParams = data["modelParams"]
    predictedFieldName = data["predictedFieldName"]

    if name in g_models.keys():
      raise web.badrequest("Model with name <%s> already exists" % name)

    model = ModelFactory.create(modelParams)
    model.enableInference({'predictedField': predictedFieldName})
    g_models[name] = model

    return json.dumps({"success": name})



class ModelRunner(object):

  def POST(self, name):
    """
    /models/{name}/run

    schema:
      {
        predictedFieldName: value
        timestamp: %m/%d/%y %H:%M
      }
      NOTE: predictedFieldName MUST be the same name specified when
            creating the model.

    returns:
    {
      "predictionNumber":<number of record>,
      "anomalyScore":anomalyScore
    }
    """
    global g_models

    data = json.loads(web.data())
    data["timestamp"] = datetime.datetime.strptime(
        data["timestamp"], "%m/%d/%y %H:%M")

    if name not in g_models.keys():
      raise web.notfound("Model with name <%s> does not exist." % name)

    modelResult = g_models[name].run(data)
    predictionNumber = modelResult.predictionNumber
    anomalyScore = modelResult.inferences["anomalyScore"]

    return json.dumps({"predictionNumber": predictionNumber,
                       "anomalyScore": anomalyScore})



web.config.debug = False
app = web.application(urls, globals())



if __name__ == "__main__":
  app.run()



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2014, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

import unittest

from nupic.data.generators.sequence_machine import SequenceMachine
from nupic.research.monitor_mixin.temporal_memory_monitor_mixin import (
  TemporalMemoryMonitorMixin)
from nupic.research.temporal_memory import TemporalMemory
# Uncomment the lines below to run tests with TP10X2 implementation instead
# from nupic.research.temporal_memory_shim import (
#   TemporalMemoryShim as TemporalMemory)
# Uncomment the lines below to run tests with FastTemporalMemory implementation
# from nupic.research.fast_temporal_memory import (
#   FastTemporalMemory as TemporalMemory)
class MonitoredTemporalMemory(TemporalMemoryMonitorMixin, TemporalMemory): pass



class AbstractTemporalMemoryTest(unittest.TestCase):

  VERBOSITY = 1
  DEFAULT_TM_PARAMS = {}
  PATTERN_MACHINE = None


  def setUp(self):
    self.tm = None
    self.patternMachine = None
    self.sequenceMachine = None


  def init(self, overrides=None):
    """
    Initialize Temporal Memory, and other member variables.

    :param overrides: overrides for default Temporal Memory parameters
    """
    params = self._computeTMParams(overrides)
    self.tm = MonitoredTemporalMemory(**params)

    self.patternMachine = self.PATTERN_MACHINE
    self.sequenceMachine = SequenceMachine(self.patternMachine)


  def feedTM(self, sequence, learn=True, num=1):
    repeatedSequence = sequence * num

    self.tm.mmClearHistory()

    for pattern in repeatedSequence:
      if pattern is None:
        self.tm.reset()
      else:
        self.tm.compute(pattern, learn=learn)


  # ==============================
  # Helper functions
  # ==============================

  def _computeTMParams(self, overrides):
    params = dict(self.DEFAULT_TM_PARAMS)
    params.update(overrides or {})
    return params




# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2014-2015, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

"""
Plot class used in monitor mixin framework.
"""

import logging
import os

try:
  # We import in here to avoid creating a matplotlib dependency in nupic.
  import matplotlib.pyplot as plt
  import matplotlib.cm as cm
except ImportError:
  # Suppress; we log it at debug level to avoid polluting the logs of apps
  # and services that don't care about plotting
  logging.debug("Cannot import matplotlib. Plot class will not work.",
                exc_info=True)



class Plot(object):


  def __init__(self, monitor, title, show=True):
    """

    @param monitor (MonitorMixinBase) Monitor Mixin instance that generated
                                      this plot

    @param title  (string)            Plot title
    """
    self._monitor = monitor
    self._title = title
    self._fig = self._initFigure()
    self._show = show
    if self._show:
      plt.ion()
      plt.show()


  def _initFigure(self):
    fig = plt.figure()
    fig.suptitle(self._prettyPrintTitle())
    return fig


  def _prettyPrintTitle(self):
    if self._monitor.mmName is not None:
      return "[{0}] {1}".format(self._monitor.mmName, self._title)
    return self._title


  def addGraph(self, data, position=111, xlabel=None, ylabel=None):
    """ Adds a graph to the plot's figure.

    @param data See matplotlib.Axes.plot documentation.
    @param position A 3-digit number. The first two digits define a 2D grid
            where subplots may be added. The final digit specifies the nth grid
            location for the added subplot
    @param xlabel text to be displayed on the x-axis
    @param ylabel text to be displayed on the y-axis
    """
    ax = self._addBase(position, xlabel=xlabel, ylabel=ylabel)
    ax.plot(data)
    plt.draw()


  def addHistogram(self, data, position=111, xlabel=None, ylabel=None,
                   bins=None):
    """ Adds a histogram to the plot's figure.

    @param data See matplotlib.Axes.hist documentation.
    @param position A 3-digit number. The first two digits define a 2D grid
            where subplots may be added. The final digit specifies the nth grid
            location for the added subplot
    @param xlabel text to be displayed on the x-axis
    @param ylabel text to be displayed on the y-axis
    """
    ax = self._addBase(position, xlabel=xlabel, ylabel=ylabel)
    ax.hist(data, bins=bins, color="green", alpha=0.8)
    plt.draw()


  def add2DArray(self, data, position=111, xlabel=None, ylabel=None, cmap=None,
                 aspect="auto", interpolation="nearest", name=None):
    """ Adds an image to the plot's figure.

    @param data a 2D array. See matplotlib.Axes.imshow documentation.
    @param position A 3-digit number. The first two digits define a 2D grid
            where subplots may be added. The final digit specifies the nth grid
            location for the added subplot
    @param xlabel text to be displayed on the x-axis
    @param ylabel text to be displayed on the y-axis
    @param cmap color map used in the rendering
    @param aspect how aspect ratio is handled during resize
    @param interpolation interpolation method
    """
    if cmap is None:
      # The default colormodel is an ugly blue-red model.
      cmap = cm.Greys

    ax = self._addBase(position, xlabel=xlabel, ylabel=ylabel)
    ax.imshow(data, cmap=cmap, aspect=aspect, interpolation=interpolation)

    if self._show:
      plt.draw()

    if name is not None:
      if not os.path.exists("log"):
        os.mkdir("log")
      plt.savefig("log/{name}.png".format(name=name), bbox_inches="tight",
                  figsize=(8, 6), dpi=400)



  def _addBase(self, position, xlabel=None, ylabel=None):
    """ Adds a subplot to the plot's figure at specified position.

    @param position A 3-digit number. The first two digits define a 2D grid
            where subplots may be added. The final digit specifies the nth grid
            location for the added subplot
    @param xlabel text to be displayed on the x-axis
    @param ylabel text to be displayed on the y-axis
    @returns (matplotlib.Axes) Axes instance
    """
    ax = self._fig.add_subplot(position)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    return ax



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

import math

import numpy
from nupic.data import SENTINEL_VALUE_FOR_MISSING_DATA
from nupic.data.fieldmeta import FieldMetaType
from nupic.encoders.base import Encoder, EncoderResult
from nupic.encoders.scalar import ScalarEncoder


class LogEncoder(Encoder):
  """
  This class wraps the ScalarEncoder class.

  A Log encoder represents a floating point value on a logarithmic scale.

  valueToEncode = log10(input)

    w -- number of bits to set in output
    minval -- minimum input value. must be greater than 0. Lower values are
              reset to this value
    maxval -- maximum input value (input is strictly less if periodic == True)
    periodic -- If true, then the input value "wraps around" such that minval =
              maxval For a periodic value, the input must be strictly less than
              maxval, otherwise maxval is a true upper bound.

    Exactly one of n, radius, resolution must be set. "0" is a special
    value that means "not set".

    n -- number of bits in the representation (must be > w)
    radius -- inputs separated by more than this distance in log space will have
              non-overlapping representations
    resolution -- The minimum change in scaled value needed to produce a change
                  in encoding. This should be specified in log space. For
                  example, the scaled values 10 and 11 will be distinguishable
                  in the output. In terms of the original input values, this
                  means 10^1 (1) and 10^1.1 (1.25) will be distinguishable.
    name -- an optional string which will become part of the description
    verbosity -- level of debugging output you want the encoder to provide.
    clipInput -- if true, non-periodic inputs smaller than minval or greater
                  than maxval will be clipped to minval/maxval
    forced -- (default False), if True, skip some safety checks

  """

  def __init__(self,
               w=5,
               minval=1e-07,
               maxval=10000,
               periodic=False,
               n=0,
               radius=0,
               resolution=0,
               name="log",
               verbosity=0,
               clipInput=True,
               forced=False):

    # Lower bound for log encoding near machine precision limit
    lowLimit = 1e-07

    # Limit minval as log10(0) is undefined.
    if minval < lowLimit:
      minval = lowLimit

    # Check that minval is still lower than maxval
    if not minval < maxval:
      raise ValueError("Max val must be larger than min val or the lower limit "
                       "for this encoder %.7f" % lowLimit)

    self.encoders = None
    self.verbosity = verbosity

    # Scale values for calculations within the class
    self.minScaledValue = math.log10(minval)
    self.maxScaledValue = math.log10(maxval)

    if not self.maxScaledValue > self.minScaledValue:
      raise ValueError("Max val must be larger, in log space, than min val.")

    self.clipInput = clipInput
    self.minval = minval
    self.maxval = maxval

    self.encoder = ScalarEncoder(w=w,
                                 minval=self.minScaledValue,
                                 maxval=self.maxScaledValue,
                                 periodic=False,
                                 n=n,
                                 radius=radius,
                                 resolution=resolution,
                                 verbosity=self.verbosity,
                                 clipInput=self.clipInput,
				 forced=forced)
    self.width = self.encoder.getWidth()
    self.description = [(name, 0)]
    self.name = name

    # This list is created by getBucketValues() the first time it is called,
    #  and re-created whenever our buckets would be re-arranged.
    self._bucketValues = None


  def getWidth(self):
    return self.width


  def getDescription(self):
    return self.description


  def getDecoderOutputFieldTypes(self):
    """
    Encoder class virtual method override
    """
    return (FieldMetaType.float, )


  def _getScaledValue(self, inpt):
    """
    Convert the input, which is in normal space, into log space
    """
    if inpt == SENTINEL_VALUE_FOR_MISSING_DATA:
      return None
    else:
      val = inpt
      if val < self.minval:
        val = self.minval
      elif val > self.maxval:
        val = self.maxval

      scaledVal = math.log10(val)
      return scaledVal


  def getBucketIndices(self, inpt):
    """
    See the function description in base.py
    """

    # Get the scaled value
    scaledVal = self._getScaledValue(inpt)

    if scaledVal is None:
      return [None]
    else:
      return self.encoder.getBucketIndices(scaledVal)


  def encodeIntoArray(self, inpt, output):
    """
    See the function description in base.py
    """

    # Get the scaled value
    scaledVal = self._getScaledValue(inpt)

    if scaledVal is None:
      output[0:] = 0
    else:
      self.encoder.encodeIntoArray(scaledVal, output)

      if self.verbosity >= 2:
        print "input:", inpt, "scaledVal:", scaledVal, "output:", output
        print "decoded:", self.decodedToStr(self.decode(output))


  def decode(self, encoded, parentFieldName=''):
    """
    See the function description in base.py
    """

    # Get the scalar values from the underlying scalar encoder
    (fieldsDict, fieldNames) = self.encoder.decode(encoded)
    if len(fieldsDict) == 0:
      return (fieldsDict, fieldNames)

    # Expect only 1 field
    assert(len(fieldsDict) == 1)

    # Convert each range into normal space
    (inRanges, inDesc) = fieldsDict.values()[0]
    outRanges = []
    for (minV, maxV) in inRanges:
      outRanges.append((math.pow(10, minV),
                        math.pow(10, maxV)))

    # Generate a text description of the ranges
    desc = ""
    numRanges = len(outRanges)
    for i in xrange(numRanges):
      if outRanges[i][0] != outRanges[i][1]:
        desc += "%.2f-%.2f" % (outRanges[i][0], outRanges[i][1])
      else:
        desc += "%.2f" % (outRanges[i][0])
      if i < numRanges-1:
        desc += ", "

    # Return result
    if parentFieldName != '':
      fieldName = "%s.%s" % (parentFieldName, self.name)
    else:
      fieldName = self.name
    return ({fieldName: (outRanges, desc)}, [fieldName])


  def getBucketValues(self):
    """
    See the function description in base.py
    """

    # Need to re-create?
    if self._bucketValues is None:
      scaledValues = self.encoder.getBucketValues()
      self._bucketValues = []
      for scaledValue in scaledValues:
        value = math.pow(10, scaledValue)
        self._bucketValues.append(value)

    return self._bucketValues


  def getBucketInfo(self, buckets):
    """
    See the function description in base.py
    """

    scaledResult = self.encoder.getBucketInfo(buckets)[0]
    scaledValue = scaledResult.value
    value = math.pow(10, scaledValue)

    return [EncoderResult(value=value, scalar=value,
                         encoding = scaledResult.encoding)]


  def topDownCompute(self, encoded):
    """
    See the function description in base.py
    """

    scaledResult = self.encoder.topDownCompute(encoded)[0]
    scaledValue = scaledResult.value
    value = math.pow(10, scaledValue)

    return EncoderResult(value=value, scalar=value,
                         encoding = scaledResult.encoding)


  def closenessScores(self, expValues, actValues, fractional=True):
    """
    See the function description in base.py
    """

    # Compute the percent error in log space
    if expValues[0] > 0:
      expValue = math.log10(expValues[0])
    else:
      expValue = self.minScaledValue

    if actValues  [0] > 0:
      actValue = math.log10(actValues[0])
    else:
      actValue = self.minScaledValue

    if fractional:
      err = abs(expValue - actValue)
      pctErr = err / (self.maxScaledValue - self.minScaledValue)
      pctErr = min(1.0, pctErr)
      closeness = 1.0 - pctErr
    else:
      err = abs(expValue - actValue)
      closeness = err

    #print "log::", "expValue:", expValues[0], "actValue:", actValues[0], \
    #      "closeness", closeness
    #import pdb; pdb.set_trace()
    return numpy.array([closeness])


  @classmethod
  def read(cls, proto):
    encoder = object.__new__(cls)
    encoder.verbosity = proto.verbosity
    encoder.minScaledValue = proto.minScaledValue
    encoder.maxScaledValue = proto.maxScaledValue
    encoder.clipInput = proto.clipInput
    encoder.minval = proto.minval
    encoder.maxval = proto.maxval
    encoder.encoder = ScalarEncoder.read(proto.encoder)
    encoder.name = proto.name
    encoder.width = encoder.encoder.getWidth()
    encoder.description = [(encoder.name, 0)]
    encoder._bucketValues = None
    return encoder


  def write(self, proto):
    proto.verbosity = self.verbosity
    proto.minScaledValue = self.minScaledValue
    proto.maxScaledValue = self.maxScaledValue
    proto.clipInput = self.clipInput
    proto.minval = self.minval
    proto.maxval = self.maxval
    self.encoder.write(proto.encoder)
    proto.name = self.name
